"""Batteries-included ``sqlalchemy`` queries for SQLite3 database."""

from collections.abc import Iterator
import functools
import logging

import sqlalchemy as sa
from sqlalchemy import select
from sqlalchemy.orm import aliased

from . import _globals
from . import _tools
from . import backend as _backend
from .backend import views as _views
from .models import (LEVEL, FAMILY, LANGUAGE, DIALECT,
                     SPECIAL_FAMILIES, BOOKKEEPING,
                     ALTNAME_PROVIDER, IDENTIFIER_SITE,
                     Languoid,
                     languoid_macroarea,
                     languoid_country, Country,
                     Link, Source, SourceProvider, Timespan, Bibfile, Bibitem,
                     Altname, AltnameProvider, Trigger,
                     Identifier, IdentifierSite,
                     ClassificationComment, ClassificationRef,
                     Endangerment, EndangermentSource,
                     EthnologueComment,
                     IsoRetirement, IsoRetirementChangeTo)

__all__ = ['get_stats_query',
           'get_example_query',
           'get_json_query',
           'iterdescendants']


log = logging.getLogger(__name__)


@_views.register_view('stats')
def get_stats_query():
    # cf. https://glottolog.org/glottolog/glottologinformation

    def languoid_count(kind, cls=Languoid, fromclause=Languoid,
                       level=None, is_root=None):
        select_nrows = (select(sa.literal(kind).label('kind'),
                               sa.func.count().label('n'))
                        .select_from(fromclause))

        if level is not None:
            select_nrows = select_nrows.where(cls.level == level)

        if is_root is not None:
            cond = ((cls.parent == sa.null()) if is_root
                    else (cls.parent != sa.null()))
            select_nrows = select_nrows.where(cond)

        return select_nrows

    Root, Child, root_child = Languoid.parent_descendant(innerjoin='reflexive',  # noqa: N806
                                                         parent_root=True)

    language_count = functools.partial(languoid_count,
                                       cls=Child, fromclause=root_child,
                                       level=LANGUAGE)

    def iterselects():
        yield languoid_count('languoids')

        yield languoid_count('families', level=FAMILY, is_root=True)

        yield languoid_count('isolates', level=LANGUAGE, is_root=True)

        yield languoid_count('roots', is_root=True)

        yield languoid_count('languages', level=LANGUAGE)

        yield languoid_count('subfamilies', level=FAMILY, is_root=False)

        yield languoid_count('dialects', level=DIALECT)

        yield (language_count('Spoken L1 Languages')
               .where(Root.name.notin_(SPECIAL_FAMILIES + (BOOKKEEPING,))))

        for name in SPECIAL_FAMILIES:
            yield language_count(name).where(Root.name == name)

        yield language_count('All').where(Root.name != BOOKKEEPING)

        yield language_count(BOOKKEEPING).where(Root.name == BOOKKEEPING)

    return sa.union_all(*iterselects())


@_views.register_view('example')
def get_example_query(*, order_by: str = 'id') -> sa.sql.Select:
    """Return example sqlalchemy core query (one denormalized row per languoid)."""
    path, family, language = Languoid.path_family_language()

    select_languoid = (select(Languoid.id,
                              Languoid.name,
                              Languoid.level,
                              Languoid.parent_id,
                              path.label('path'),
                              family.label('family_id'),
                              language.label('dialect_language_id'),
                              Languoid.hid,
                              Languoid.iso639_3,
                              Languoid.latitude,
                              Languoid.longitude,
                              select_languoid_macroareas(as_json=False),
                              select_languoid_countries(as_json=False),
                              select_languoid_links(as_json=False),
                              select_languoid_sources(provider_name='glottolog',
                                                      as_json=False))
                       .select_from(Languoid))

    for provider_name in sorted(ALTNAME_PROVIDER):
        altnames = select_languoid_altnames(provider_name=provider_name,
                                            label='altnames_{provider_name}',
                                            as_json=False)
        select_languoid = select_languoid.add_columns(altnames)

    for field_name in ('lgcode', 'inlg'):
        triggers = select_languoid_triggers(field_name=field_name,
                                            label='triggers_{field_name}',
                                            as_json=False)
        select_languoid = select_languoid.add_columns(triggers)

    for site_name in sorted(IDENTIFIER_SITE):
        select_languoid = add_identifier(select_languoid, site_name,
                                         label='identifier_{site_name}')

    for kind in ('sub', 'family'):
        select_languoid = add_classification_comment(select_languoid, kind,
                                                     label='classification_{kind}')
        select_languoid = add_classification_refs(select_languoid, kind,
                                                  label='classification_{kind}refs')

    select_languoid = add_model_columns(select_languoid, Endangerment,
                                        label='endangerment_{name}')

    select_languoid = add_endangermentsource(select_languoid,
                                             label='endangerment_source')

    select_languoid = add_model_columns(select_languoid, EthnologueComment,
                                        label='elcomment_{name}',
                                        add_outerjoin=EthnologueComment)

    select_languoid = add_model_columns(select_languoid, IsoRetirement,
                                        label='iso_retirement_{name}',
                                        add_outerjoin=IsoRetirement)

    change_to = select_iso_retirement_change_to(label='iso_retirement_change_to',
                                                as_json=False)
    select_languoid = select_languoid.add_columns(change_to)

    select_languoid = add_order_by(select_languoid,
                                   order_by=order_by,
                                   column_for_path_order=path)

    return select_languoid


def add_order_by(select_languoid: sa.sql.Select, /, *,
                 order_by: str, column_for_path_order) -> sa.sql.Select:
    if order_by in (True, None, 'id'):
        return select_languoid.order_by(Languoid.id)
    elif order_by == 'path':
        return select_languoid.order_by(column_for_path_order)
    elif order_by is False:  # pragma: no cover
        return select_languoid
    raise ValueError(f'{order_by=!r} not implemented')  # pragma: no cover


def add_model_columns(select_languoid: sa.sql.Select, model, /, *,
                      add_outerjoin=None, label: str = '{name}',
                      ignore: str = 'id') -> sa.sql.Select:
    columns = model.__table__.columns
    if ignore:
        ignore_suffix = f'_{ignore}'
        columns = [c for c in columns
                   if c.name != ignore and not c.name.endswith(ignore_suffix)]

    columns = [c.label(label.format(name=c.name)) for c in columns]
    select_languoid = select_languoid.add_columns(*columns)
    if add_outerjoin is not None:
        select_languoid = select_languoid.outerjoin(add_outerjoin)
    return select_languoid


def group_concat(x, /, *, separator: str = ', '):
    return sa.func.group_concat(x, separator)


def add_identifier(select_languoid: sa.sql.Select, site_name: str, /, *,
                   label: str) -> sa.sql.Select:
    identifier = aliased(Identifier, name=f'ident_{site_name}')
    site = aliased(IdentifierSite, name=f'ident_{site_name}_site')
    label = label.format(site_name=site_name)

    return (select_languoid.add_columns(identifier.identifier.label(label))
            .outerjoin(sa.join(identifier, site, identifier.site_id == site.id),
                       sa.and_(site.name == site_name,
                               identifier.languoid_id == Languoid.id)))


def add_classification_comment(select_languoid: sa.sql.Select, kind: str, /, *,
                               label: str,
                               bib_suffix: str = '_cr') -> sa.sql.Select:
    comment = aliased(ClassificationComment, name=f'cc_{kind}')
    label = label.format(kind=kind)

    return (select_languoid.add_columns(comment.comment.label(label))
            .outerjoin(comment, sa.and_(comment.kind == kind,
                                        comment.languoid_id == Languoid.id)))


def add_classification_refs(select_languoid: sa.sql.Select, kind: str, /, *,
                            label: str,
                            bib_suffix: str = '_cr') -> sa.sql.Select:
    ref = aliased(ClassificationRef, name=f'cr_{kind}')
    bibfile = aliased(Bibfile, name=f'bibfile{bib_suffix}_{kind}')
    bibitem = aliased(Bibitem, name=f'bibitem{bib_suffix}_{kind}')
    label = label.format(kind=kind)

    ref = (select(ref.printf(bibfile, bibitem))
           .select_from(ref)
           .filter_by(languoid_id=Languoid.id)
           .correlate(Languoid)
           .filter_by(kind=kind)
           .join(ref.bibitem.of_type(bibitem))
           .join(bibitem.bibfile.of_type(bibfile))
           .order_by(ref.ord)
           .alias(f'lang_cref_{kind}'))

    refs = select(group_concat(ref.c.printf).label(label)).label(label)
    return select_languoid.add_columns(refs)


def add_endangermentsource(select_languoid: sa.sql.Select, /, *,
                           label: str,
                           bib_suffix: str = '_e') -> sa.sql.Select:
    bibfile = aliased(Bibfile, name=f'bibfile{bib_suffix}')
    bibitem = aliased(Bibitem, name=f'bibitem{bib_suffix}')

    endangermentsource = (EndangermentSource.printf(bibfile, bibitem)
                          .label(label))

    return (select_languoid.add_columns(endangermentsource)
            .outerjoin(sa.join(Endangerment, EndangermentSource))
            .outerjoin(sa.join(bibitem, bibfile)))


group_array = sa.func.json_group_array


group_object = sa.func.json_group_object


@_views.register_view('path_languoid', as_rows=True, load_json=False)
def get_json_query(*, limit: int | None = None,
                   offset: int | None = 0,
                   order_by: str = _globals.LANGUOID_ORDER,
                   as_rows: bool = False,
                   load_json: bool = True,
                   sort_keys: bool = False,
                   path_label: str = _globals.PATH_LABEL,
                   languoid_label: str = _globals.LANGUOID_LABEL) -> sa.sql.Select:
    languoid = {'id': Languoid.id,
                'parent_id': Languoid.parent_id,
                'name': Languoid.name,
                'level': Languoid.level,
                'hid': Languoid.hid,
                'iso639_3': Languoid.iso639_3,
                'latitude': Languoid.latitude,
                'longitude': Languoid.longitude,
                'macroareas': select_languoid_macroareas(as_json=True),
                'countries': select_languoid_countries(as_json=True, sort_keys=sort_keys),
                'links': select_languoid_links(as_json=True, sort_keys=sort_keys),
                'timespan': select_languoid_timespan(as_json=True, sort_keys=sort_keys),
                'sources': select_languoid_sources(as_json=True, sort_keys=sort_keys),
                'altnames': select_languoid_altnames(as_json=True, sort_keys=sort_keys),
                'triggers': select_languoid_triggers(as_json=True),
                'identifier': select_languoid_identifier(),
                'classification': select_languoid_classification(sort_keys=sort_keys),
                'endangerment': select_languoid_endangerment(sort_keys=sort_keys),
                'hh_ethnologue_comment': select_languoid_hh_ethnologue_comment(sort_keys=sort_keys),
                'iso_retirement': select_languoid_iso_retirement(sort_keys=sort_keys)}

    json_object = functools.partial(_backend.json_object,
                                    sort_keys_=sort_keys,
                                    load_json_=load_json)
    del sort_keys, load_json

    if as_rows:
        path = column_for_path_order = Languoid.path()
        columns = [path.label(path_label),
                   json_object(label_=languoid_label, **languoid)]
    else:
        subquery = Languoid._path_part(include_self=True, bottomup=False)

        path_array = (sa.func.json_group_array(subquery.c.path_part)
                     .label('path_array'))
        path_array = select(path_array).label('path')

        columns = [json_object(label_=_globals.LANGUOID_FILE_BASENAME,
                               **{path_label: path_array,
                                  languoid_label: json_object(**languoid)})]

        column_for_path_order = (sa.func.group_concat(subquery.c.path_part,
                                                      _globals.FILE_PATH_SEP)
                                 .label('path_string'))
        column_for_path_order = select(column_for_path_order).label('file_path')

    select_json = select(*columns).select_from(Languoid)
    select_json = add_order_by(select_json,
                               order_by=order_by,
                               column_for_path_order=column_for_path_order)

    if offset:
        select_json = select_json.offset(offset)
    if limit is not None:
        select_json = select_json.limit(limit)
    return select_json


def select_languoid_macroareas(languoid=Languoid, /, *, as_json: bool,
                               label: str = 'macroareas',
                               alias: str = 'lang_ma') -> sa.sql.Select:
    name = languoid_macroarea.c.macroarea_name

    macroarea = (select(name)
                 .select_from(languoid_macroarea)
                 .filter_by(languoid_id=languoid.id)
                 .correlate(languoid)
                 .order_by(name)
                 .alias(alias))

    aggregate = group_array if as_json else group_concat

    macroareas = aggregate(macroarea.c.macroarea_name)

    return select(macroareas.label(label)).label(label)


def select_languoid_countries(languoid=Languoid, /, *, as_json: bool,
                              label: str = 'countries',
                              sort_keys: bool = False,
                              alias: str = 'lang_country') -> sa.sql.Select:
    value = (Country.jsonf(sort_keys=sort_keys) if as_json else
             languoid_country.c.country_id)

    country = (select(value)
               .select_from(languoid_country)
               .filter_by(languoid_id=languoid.id)
               .correlate(languoid))

    if as_json:
        country = (country.join(Country)
                   .order_by(Country.printf())
                   .alias(alias))

        countries = group_array(sa.func.json(country.c.jsonf))
    else:
        country = (country
                   .order_by(value)
                   .alias(alias))

        countries = group_concat(country.c.country_id)

    return select(countries.label(label)).label(label)


def select_languoid_links(languoid=Languoid, *, as_json: bool,
                          label: str = 'links',
                          sort_keys: bool = False,
                          alias: str = 'lang_link') -> sa.sql.Select:
    link = (select(Link.jsonf(sort_keys=sort_keys) if as_json else
                   Link.printf())
            .select_from(Link)
            .filter_by(languoid_id=languoid.id)
            .correlate(languoid)
            .order_by(Link.ord)
            .alias(alias))

    links = group_array(sa.func.json(link.c.jsonf)) if as_json else group_concat(link.c.printf)

    return select(links.label(label)).label(label)


def select_languoid_timespan(languoid=Languoid, /, *, as_json: bool,
                             label: str = 'timespan',
                             sort_keys: bool = False) -> sa.sql.Select:
    return (select(Timespan.jsonf(sort_keys=sort_keys) if as_json else
                   Timespan.printf())
            .select_from(Timespan)
            .filter_by(languoid_id=languoid.id)
            .correlate(languoid)
            .label(label))


def select_languoid_sources(languoid=Languoid, /, *, as_json: bool,
                            provider_name: str | None = None,
                            label: str = 'sources',
                            sort_keys: bool = False,
                            alias: str = 'lang_source',
                            bib_prefix: str = 'source_') -> sa.sql.Select:
    source = (aliased(Source, name=f'source_{provider_name}')
              if provider_name is not None else Source)

    provider = aliased(SourceProvider, name='source_provider')

    bibitem = aliased(Bibitem, name=f'{bib_prefix}bibitem')
    bibfile = aliased(Bibfile, name=f'{bib_prefix}bibfile')

    columns = [source.jsonf(bibfile, bibitem, sort_keys=sort_keys)
               if as_json else source.printf(bibfile, bibitem)]

    order_by = [bibfile.name, bibitem.bibkey]

    if provider_name is None:
        name = provider.name
        columns.insert(0, name.label('provider'))
        order_by.insert(0, name)

    source = (select(*columns)
              .select_from(source)
              .filter_by(languoid_id=languoid.id)
              .correlate(languoid)
              .join(Source.provider.of_type(provider))
              .join(Source.bibitem.of_type(bibitem))
              .join(bibitem.bibfile.of_type(bibfile))
              .order_by(*order_by))

    if provider_name is not None:
        source = source.where(provider.name == provider_name)
        alias = f'{alias}_{provider_name}'

    source = source.alias(alias)

    sub_label = f'{label}_{provider_name}' if provider_name else label

    value = (group_array(sa.func.json(source.c.jsonf)).label('value')
             if as_json else group_concat(source.c.printf).label(sub_label))

    if provider_name is not None:
        return select(value).label(sub_label)

    key = source.c.provider
    sources = (select(key.label('key'), value)
               .group_by(key)
               .alias(alias))

    if as_json:
        sources = sa.func.nullif(group_object(sources.c.key,
                                              sa.func.json(sources.c.value)),
                                 '{}')
    else:  # pragma: no cover
        raise NotImplementedError

    return select(sources.label(label)).label(label)


def select_languoid_altnames(languoid=Languoid, /, *, as_json: bool,
                             provider_name: str | None = None,
                             label: str = 'altnames',
                             sort_keys: bool = False,
                             alias: str = 'lang_altname') -> sa.sql.Select:
    if provider_name is not None:
        altname = aliased(Altname, name=f'altname_{provider_name}')
        provider = aliased(AltnameProvider, name=f'altname_{provider_name}_provider')
        columns = []
        order_by = [altname.name, altname.lang]
        label = label.format(provider_name=provider_name)
        alias = f'{alias}_{provider_name}'
    else:
        altname = Altname
        provider = aliased(AltnameProvider, name='altname_provider')
        columns = [provider.name.label('provider')]
        order_by = [provider.name, altname.printf()]

    columns.append(altname.jsonf(sort_keys=sort_keys) if as_json
                   else altname.printf())

    altname = (select(*columns)
               .select_from(altname)
               .filter_by(languoid_id=languoid.id)
               .correlate(languoid)
               .join(altname.provider.of_type(provider))
               .order_by(*order_by))

    if provider_name is not None:
        altname = altname.where(provider.name == provider_name)

    altname = altname.alias(alias)

    value = (group_array(sa.func.json(altname.c.jsonf)).label('value')
             if as_json else group_concat(altname.c.printf).label(label))

    if provider_name is not None:
        return select(value).label(label)

    key = altname.c.provider
    altnames = (select(key.label('key'), value)
                .group_by(key)
                .alias(alias))

    if as_json:
        altnames = sa.func.nullif(group_object(altnames.c.key,
                                               sa.func.json(altnames.c.value)),
                                  '{}')
    else:  # pragma: no cover
        raise NotImplementedError

    return select(altnames.label(label)).label(label)


def select_languoid_triggers(languoid=Languoid, /, *, as_json: bool,
                             field_name: str | None = None,
                             label: str = 'triggers',
                             alias: str = 'lang_trigger') -> sa.sql.Select:
    if field_name is not None:
        trigger = aliased(Trigger, name=f'trigger_{field_name}')
        columns = [trigger.trigger]
        order_by = [trigger.ord]
        label = label.format(field_name=field_name)
        alias = f'{alias}_{field_name}'
    else:
        trigger = Trigger
        columns = [trigger.field, trigger.trigger]
        order_by = [trigger.field, trigger.ord]

    trigger = (select(*columns)
               .select_from(trigger)
               .filter_by(languoid_id=languoid.id)
               .correlate(languoid)
               .order_by(*order_by))

    if field_name is not None:
        trigger = trigger.filter_by(field=field_name)

    trigger = trigger.alias(alias)

    value = (group_array(trigger.c.trigger).label('value')
             if as_json else group_concat(trigger.c.trigger).label(label))

    if field_name is not None:
        return select(value).label(label)

    key = trigger.c.field
    triggers = (select(key.label('key'), value)
                .group_by(key)
                .alias('lang_triggers'))

    if as_json:
        triggers = sa.func.nullif(group_object(triggers.c.key,
                                               sa.func.json(triggers.c.value)),
                                  '{}')
    else:  # pragma no cover
        raise NotImplementedError

    return select(triggers.label(label)).label(label)


def select_languoid_identifier(languoid=Languoid, /, *,
                               label: str = 'identifiers') -> sa.sql.Select:
    identifier = (select(IdentifierSite.name.label('site'),
                         Identifier.identifier.label('identifier'))
                  .select_from(Identifier)
                  .filter_by(languoid_id=languoid.id)
                  .correlate(languoid)
                  .join(Identifier.site.of_type(IdentifierSite))
                  .alias('lang_identifiers'))

    site = identifier.c.site

    identifier = (select(site, identifier.c.identifier)
                  .order_by(site)
                  .alias('lang_identifiers_ordered'))

    identifier = sa.func.nullif(group_object(identifier.c.site,
                                             identifier.c.identifier),
                                '{}')

    return select(identifier.label(label)).label(label)


def select_languoid_classification(languoid=Languoid, /, *,
                                   label: str = 'classification',
                                   sort_keys: bool = False,
                                   bib_suffix: str = '_cr') -> sa.sql.Select:
    classification_comment = (select(ClassificationComment.kind.label('key'),
                                     sa.func.json_quote(ClassificationComment.comment).label('value'))
                              .select_from(ClassificationComment)
                              .filter_by(languoid_id=languoid.id)
                              .correlate(languoid)
                              .scalar_subquery())

    bibitem = aliased(Bibitem, name=f'bibitem{bib_suffix}')
    bibfile = aliased(Bibfile, name=f'bibfile{bib_suffix}')

    kind = ClassificationRef.kind

    classification_ref = (select((kind + 'refs').label('key'),
                                 ClassificationRef.jsonf(bibfile, bibitem,
                                                         sort_keys=sort_keys))
                          .select_from(ClassificationRef)
                          .filter_by(languoid_id=languoid.id)
                          .correlate(languoid)
                          .join(ClassificationRef.bibitem.of_type(bibitem))
                          .join(bibitem.bibfile.of_type(bibfile))
                          .order_by(kind, ClassificationRef.ord)
                          .alias('lang_cref'))

    classification_refs = (select(classification_ref.c.key,
                                  group_array(sa.func.json(classification_ref.c.jsonf))
                                  .label('value'))
                           .group_by(classification_ref.c.key))

    classification = (classification_comment
                      .union_all(classification_refs)
                      .alias('lang_classifciation'))

    key = classification.c.key

    classification = (select(key, classification.c.value.label('value'))
                      .order_by(key)
                      .alias('classification_object'))

    nullification = sa.func.nullif(group_object(classification.c.key,
                                                sa.func.json(classification.c.value)),
                                   '{}')

    return (select(nullification.label(label))
            .select_from(classification)
            .label(label))


def select_languoid_endangerment(languoid=Languoid, /, *,
                                 label: str = 'endangerment',
                                 sort_keys: bool = False,
                                 bib_suffix: str = '_e') -> sa.sql.Select:
    bibitem = aliased(Bibitem, name=f'bibitem{bib_suffix}')
    bibfile = aliased(Bibfile, name=f'bibfile{bib_suffix}')

    return (select(Endangerment.jsonf(EndangermentSource,
                                      bibfile, bibitem,
                                      sort_keys=sort_keys,
                                      label=label))
            .select_from(Endangerment)
            .filter_by(languoid_id=languoid.id)
            .correlate(languoid)
            .join(Endangerment.source)
            .outerjoin(sa.join(bibitem, bibfile))
            .label(label))


def select_languoid_hh_ethnologue_comment(languoid=Languoid, /, *,
                                          label: str = 'hh_ethnologue_comment',
                                          sort_keys: bool = False) -> sa.sql.Select:
    return (select(EthnologueComment
                   .jsonf(sort_keys=sort_keys, label=label))
            .select_from(EthnologueComment)
            .filter_by(languoid_id=languoid.id)
            .correlate(languoid)
            .label(label))


def select_languoid_iso_retirement(languoid=Languoid, /, *,
                                   label: str = 'iso_retirement',
                                   sort_keys: bool = False,
                                   alias: str = 'lang_irct',
                                   alias_label: str = 'change_to') -> sa.sql.Select:
    codes = select_iso_retirement_change_to(as_json=True,
                                            label=alias_label)

    return (select(IsoRetirement.jsonf(change_to=codes,
                                       sort_keys=sort_keys,
                                       optional=True,
                                       label=label))
            .select_from(IsoRetirement)
            .filter_by(languoid_id=languoid.id)
            .correlate(languoid)
            .label(label))


def select_iso_retirement_change_to(iso_retirement=IsoRetirement, /, *,
                                    as_json: bool, label: str,
                                    alias: str = 'lang_irct') -> sa.sql.Select:
    code = (select(IsoRetirementChangeTo.code)
            .select_from(IsoRetirementChangeTo)
            .filter_by(languoid_id=iso_retirement.languoid_id)
            .correlate(IsoRetirement)
            .order_by(IsoRetirementChangeTo.ord)
            .alias(alias))

    aggregate = group_array if as_json else group_concat

    codes = aggregate(code.c.code)

    return select(codes.label(label)).label(label)


def iterdescendants(parent_level: str | None = None,
                    child_level: str | None = None, *,
                    bind=_globals.ENGINE) -> Iterator[tuple[str, list[str]]]:
    """Yield pairs of (parent id, sorted list of their descendant ids)."""
    # TODO: implement ancestors/descendants as sa.orm.relationship()
    # see https://bitbucket.org/zzzeek/sqlalchemy/issues/4165
    parent_root = None
    if parent_level is None:
        pass
    elif parent_level == 'top':
        parent_root = True
        parent_level = None
    elif parent_level in LEVEL:
        pass
    else:  # pragma: no cover
        raise ValueError(f'invalid parent_level: {parent_level!r}')

    Parent, Child, parent_child = Languoid.parent_descendant(parent_root=parent_root,  # noqa: N806
                                                             parent_level=parent_level)

    select_pairs = (select(Parent.id.label('parent_id'),
                           Child.id.label('child_id'))
                    .select_from(parent_child)
                    .order_by('parent_id', 'child_id'))

    if child_level is not None:
        if child_level not in LEVEL:  # pragma: no cover
            raise ValueError(f'invalid child_level: {child_level!r}')
        select_pairs = select_pairs.where(sa.or_(Child.level == sa.null(),
                                                 Child.level == child_level))

    rows = _backend.iterrows(select_pairs, bind=bind)

    for parent_id, grp in _tools.groupby_attrgetter('parent_id')(rows):
        _, c = next(grp)
        if c is None:
            descendants = []
        else:
            descendants = [c] + [c for _, c in grp]
        yield parent_id, descendants



import sqlalchemy as sa


def test_print_versions(capsys, bare_treedb):
    assert bare_treedb.print_versions() is None

    out, err = capsys.readouterr()
    assert not err

    assert out.startswith('treedb version: ')


def test_scalar(bare_treedb):
    result = bare_treedb.scalar(sa.select(sa.func.sqlite_version()))

    assert result is not None
    assert isinstance(result, str)
    assert result



"""Raw tables schema."""

import sqlalchemy as sa

from sqlalchemy import (Column, Integer, String, Text, Boolean,
                        ForeignKey, CheckConstraint, UniqueConstraint)

from .._globals import REGISTRY as registry  # noqa: N811

__all__ = ['File', 'Option', 'Value']

PREFIX = '_'


@registry.mapped
class File:
    """Forward-slash-joined ids from the root to each item."""

    __tablename__ = f'{PREFIX}file'

    id = Column(Integer, primary_key=True)

    glottocode = Column(String(8), CheckConstraint('length(glottocode) = 8'),
                        nullable=False, unique=True)

    path = Column(Text, CheckConstraint('length(path) >= 8 AND (length(path) + 1) % 9 = 0'),
                  nullable=False, unique=True)

    size = Column(Integer, CheckConstraint('size > 0'), nullable=False)
    sha256 = Column(String(64), CheckConstraint('length(sha256) = 64'),
                    unique=True, nullable=False)

    __table_args__ = (CheckConstraint('substr(path, -length(glottocode))'
                                      ' = glottocode'),)

    @classmethod
    def path_depth(cls, /, *, label='path_depth'):
        return sa.func.floor((sa.func.length(cls.path) + 1) / 9).label(label)


@registry.mapped
class Option:
    """Unique (section, option) key of the values with lines config."""

    __tablename__ = f'{PREFIX}option'

    id = Column(Integer, primary_key=True)

    section = Column(Text, CheckConstraint("section != ''"), nullable=False)
    option = Column(Text, CheckConstraint("option != ''"), nullable=False)

    is_lines = Column(Boolean(create_constraint=True))

    defined = Column(Boolean(create_constraint=True), nullable=False)
    defined_any_options = Column(Boolean(create_constraint=True), nullable=False)

    ord_section = Column(Integer, CheckConstraint('ord_section >= 1'))
    ord_option = Column(Integer, CheckConstraint('ord_section >= 0'))

    __table_args__ = (UniqueConstraint(section, option),
                      CheckConstraint('(is_lines IS NULL) = (defined = 0)'),
                      CheckConstraint('defined = 1 OR defined_any_options = 0'),
                      CheckConstraint('(defined = 0) = (ord_section IS NULL)'),
                      CheckConstraint('ord_section IS NOT NULL'
                                      ' OR ord_option IS NULL'))


@registry.mapped
class Value:
    """Item value as (path, section, option, line, value) combination."""

    __tablename__ = f'{PREFIX}value'

    file_id = Column(ForeignKey('_file.id'), primary_key=True)
    option_id = Column(ForeignKey('_option.id'), primary_key=True)
    line = Column(Integer, CheckConstraint('line > 0'), primary_key=True)

    # TODO: consider adding version for selective updates
    value = Column(Text, CheckConstraint("value != ''"), nullable=False)

    __table_args__ = (UniqueConstraint(file_id, line),
                      {'info': {'without_rowid': True}})



"""Package-level globals."""

from collections.abc import Mapping
import datetime
from typing import NamedTuple, TypeAlias

from sqlalchemy.orm import (registry as _registry,
                            sessionmaker as _sessionmaker)

from . import _proxies

__all__ = ['DEFAULT_FILESTEM', 'CONFIG', 'MEMORY_TAG', 'DEFAULT_ROOT',
           'PATH_LABEL', 'LANGUOID_LABEL', 'LANGUOID_ORDER',
           'DEFAULT_HASH', 'FILE_PATH_SEP',
           'ENGINE', 'ROOT',
           'REGISTRY',
           'SESSION',
           'RecordItem',
           'LanguoidItem']

DEFAULT_FILESTEM = 'treedb'

CONFIG = 'treedb.ini'

LANGUOID_FILE_BASENAME = 'md.ini'

MEMORY_TAG = '-memory'

DEFAULT_ROOT = './glottolog/'

PATH_LABEL = '__path__'

LANGUOID_LABEL = 'languoid'

LANGUOID_ORDER = 'path'

DEFAULT_HASH = 'sha256'

FILE_PATH_SEP = '/'

_SQLALCHEMY_FUTURE = True

ENGINE = _proxies.SQLiteEngineProxy(future=_SQLALCHEMY_FUTURE)

ROOT = _proxies.PathProxy()

REGISTRY = _registry()

SESSION = _sessionmaker(bind=ENGINE, future=_SQLALCHEMY_FUTURE)


assert PATH_LABEL < LANGUOID_LABEL


PathType: TypeAlias = tuple[str, ...]

RecordValueType: TypeAlias = str | list[str]

RecordType: TypeAlias = Mapping[str, Mapping[str, RecordValueType]]


def filepath_tuple(file_path: str, /, *, sep=FILE_PATH_SEP) -> tuple[str]:
    path_parts = file_path.split(sep)
    return tuple(path_parts)


class RecordItem(NamedTuple):
    """Pair of path and record.

    >>> RecordItem.from_filepath_record('spam/eggs', {'core': {'id': 'abin1243'}})
    RecordItem(path=('spam', 'eggs'), record={'core': {'id': 'abin1243'}})
    """

    path: PathType

    record: RecordType

    @classmethod
    def from_filepath_record(cls, file_path: str, languoid, /):
        return cls(filepath_tuple(file_path), languoid)


ValueType: TypeAlias = (str | int | float | bool | None | datetime.datetime
                        | list[str])

MatchType: TypeAlias = Mapping[str, ValueType]

MatchList: TypeAlias = list[MatchType]

MatchListMapping: TypeAlias = Mapping[str, MatchList]

LanguoidValueType: TypeAlias = (ValueType
                                | MatchType | MatchList | MatchListMapping
                                | Mapping[str, MatchListMapping])

LanguoidType: TypeAlias = Mapping[str, LanguoidValueType]


class LanguoidItem(NamedTuple):
    """Pair of path and languoid.

    >>> LanguoidItem.from_filepath_languoid('spam/eggs', {'id': 'abin1243'})
    LanguoidItem(path=('spam', 'eggs'), languoid={'id': 'abin1243'})
    """

    path: PathType

    languoid: LanguoidType

    @classmethod
    def from_filepath_languoid(cls, file_path: str, languoid, /):
        return cls(filepath_tuple(file_path), languoid)



#!/usr/bin/env python3
# flake8: noqa

"""Run the tests with https://pytest.org."""

import pathlib
import platform
import sys

import pytest

ARGS = [#'--run-writes',
        #'--skip-slow',
        #'--skip-pandas',
        #'--skip-sqlparse',
        #'-k', 'not reads',
        #'-k', 'indent',
        #'--collect-only',
        #'--capture', 'no',  # a.k.a -s
        #'--verbose',
        #'--showlocals',  # a.k.a. -l
        #'--pdb',
        #'--exitfirst',  # a.k.a. -x
        #'-W', 'error',
        #'--loglevel-debug',
        #'--log-sql',
        #'--no-sqlalchemy-warn-20',
        #'--file-engine',
        #'--glottolog-tag', 'master',
        #'--glottolog-tag', 'v5.3',
        #'--glottolog-tag', 'v5.2.1',
        #'--glottolog-tag', 'v5.2',
        #'--glottolog-tag', 'v5.1',
        #'--glottolog-tag', 'v5.0',
        #'--glottolog-tag', 'v4.8',
        #'--glottolog-tag', 'v4.7',
        #'--glottolog-tag', 'v4.6',
        #'--glottolog-tag', 'v4.5',
        #'--glottolog-tag', 'v4.4',
        #'--glottolog-tag', 'v4.3-treedb-fixes',
        #'--glottolog-tag', 'v4.2.1',
        #'--glottolog-tag', 'v4.1',
        #'--glottolog-repo-root', './glottolog/',
        #'--rebuild', '--force-rebuild',
        #'--exclude-raw',
       ]

if platform.system() == 'Windows' and 'idlelib' in sys.modules:
    ARGS += ['-p', 'no:faulthandler']

if platform.system() == 'Windows':
    ARGS.append('--pdb')


print('run', [pathlib.Path(__file__).name] + sys.argv[1:])

if '--installed' in sys.argv[1:]:
    sys.argv[1:] = [a for a in sys.argv[1:] if a != '--installed']
    path_item = sys.path.pop(0)
    print(f'removed {path_item!r} from sys.path')
    ARGS += ['--import-mode', 'append']

args = ARGS + sys.argv[1:]

# https://docs.pytest.org/en/stable/reference/reference.html#pytest-main
print(f'pytest.main({args!r})')
if (returncode := pytest.main(args)):
    print('FAILED:', returncode)
    sys.exit(returncode)
print('PASSED.')



"""Parse a ``languoids/tree/**/md.ini`` record into a languoid and serialize it."""

from collections.abc import Iterable, Iterator
import datetime
import logging
import operator
import re
import warnings

import pycountry

from .. import _globals

from . import fields as _fields

__all__ = ['pipe']

(CORE,
 SOURCES,
 ALTNAMES, TRIGGERS, IDENTIFIER,
 CLASSIFICATION,
 ENGANGERMENT, HH_ETHNOLOGUE_COMMENT, ISO_RETIREMENT) = _fields.SECTIONS

FLOAT_DIGITS = 12

FLOAT_FORMAT = f'%.{FLOAT_DIGITS}f'

DATE_FORMAT = '%Y-%m-%d'

DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S'

ISO_8601_INTERVAL = re.compile(r'''
(?P<start_sign>[+-]?)
(?P<start_date>\d{1,4}-\d{2}-\d{2})
/
(?P<end_sign>[+-]?)
(?P<end_date>\d{1,4}-\d{2}-\d{2})
'''.strip(), flags=(re.ASCII | re.VERBOSE))


log = logging.getLogger(__name__)


def pipe(items, /, *, dump: bool,
         convert_lines: bool):
    codec = _dump if dump else _parse
    return codec(items, convert_lines=convert_lines)


def _parse(records: Iterable[_globals.RecordItem], /, *,
           convert_lines: bool) -> Iterator[_globals.LanguoidItem]:
    r"""Yield languoid items from given record ítems (from raw).

    >>> dict(pipe({('abin1243',):
    ...            {'core': {'name': 'Abinomn',
    ...                      'hid': 'bsa',
    ...                      'level': 'language',
    ...                      'iso639-3': 'bsa',
    ...                      'latitude': '-2.92281',
    ...                      'longitude': '138.891',
    ...                      'macroareas': '\nPapunesia',
    ...                      'countries': '\nID',
    ...                      'links': ('\n[Abinomn](http://endangeredlanguages.com/lang/1763)'
    ...                                '\nhttps://www.wikidata.org/entity/Q56648'
    ...                                '\nhttps://en.wikipedia.org/wiki/Abinomn_language')}}}.items(),
    ...           dump=False, convert_lines=True))  # doctest: +NORMALIZE_WHITESPACE
    {('abin1243',):
     {'id': 'abin1243',
      'parent_id': None,
      'name': 'Abinomn',
      'level': 'language',
      'hid': 'bsa',
      'iso639_3': 'bsa',
      'latitude': -2.92281,
      'longitude': 138.891,
      'macroareas': ['Papunesia'],
      'countries': [{'id': 'ID', 'name': 'Indonesia'}],
      'links': [{'url': 'http://endangeredlanguages.com/lang/1763', 'title': 'Abinomn', 'scheme': 'http'},
                {'url': 'https://www.wikidata.org/entity/Q56648', 'title': None, 'scheme': 'https'},
                {'url': 'https://en.wikipedia.org/wiki/Abinomn_language', 'title': None, 'scheme': 'https'}],
      'timespan': None,
      'sources': None,
      'altnames': None,
      'triggers': None,
      'identifier': None,
      'classification': None,
      'endangerment': None,
      'hh_ethnologue_comment': None,
      'iso_retirement': None}}

    >>> dict(pipe({('abin1243',):
    ...            {'core': {'name': 'Abinomn',
    ...                      'hid': 'bsa',
    ...                      'level': 'language',
    ...                      'iso639-3': 'bsa',
    ...                      'latitude': '-2.92281',
    ...                      'longitude': '138.891',
    ...                      'macroareas': ['Papunesia'],
    ...                      'countries': ['ID'],
    ...                      'links': ['[Abinomn](http://endangeredlanguages.com/lang/1763)',
    ...                                'https://www.wikidata.org/entity/Q56648',
    ...                                'https://en.wikipedia.org/wiki/Abinomn_language'],
    ...                      'timespan': None}}}.items(),
    ...           dump=False, convert_lines=False))  # doctest: +NORMALIZE_WHITESPACE
    {('abin1243',):
     {'id': 'abin1243',
      'parent_id': None,
      'name': 'Abinomn',
      'level': 'language',
      'hid': 'bsa',
      'iso639_3': 'bsa',
      'latitude': -2.92281,
      'longitude': 138.891,
      'macroareas': ['Papunesia'],
      'countries': [{'id': 'ID', 'name': 'Indonesia'}],
      'links': [{'url': 'http://endangeredlanguages.com/lang/1763', 'title': 'Abinomn', 'scheme': 'http'},
                {'url': 'https://www.wikidata.org/entity/Q56648', 'title': None, 'scheme': 'https'},
                {'url': 'https://en.wikipedia.org/wiki/Abinomn_language', 'title': None, 'scheme': 'https'}],
      'timespan': None,
      'sources': None,
      'altnames': None,
      'triggers': None,
      'identifier': None,
      'classification': None,
      'endangerment': None,
      'hh_ethnologue_comment': None,
      'iso_retirement': None}}
    """
    n = 0
    make_item = _globals.LanguoidItem
    for n, (path, cfg) in enumerate(records, start=1):
        languoid = make_languoid(path, cfg, convert_lines=convert_lines)
        yield make_item(path, languoid)
    log.info('%s languoids extracted from records', f'{n:_d}')


def _dump(languoids: Iterable[_globals.LanguoidItem], /, *,
          convert_lines: bool) -> Iterator[_globals.RecordItem]:
    r"""

    >>> dict(pipe({('abin1243',): {
    ...             'id': 'abin1243',
    ...             'parent_id': None,
    ...             'name': 'Abinomn',
    ...             'level': 'language',
    ...             'hid': 'bsa',
    ...             'iso639_3': 'bsa',
    ...             'latitude': -2.92281,
    ...             'longitude': 138.891,
    ...             'macroareas': ['Papunesia'],
    ...             'countries': [{'name': 'Indonesia', 'id': 'ID'}],
    ...             'links': [{'url': 'http://endangeredlanguages.com/lang/1763',
    ...                        'title': 'Abinomn', 'scheme': 'http'},
    ...                       {'url': 'https://www.wikidata.org/entity/Q56648',
    ...                         'title': None, 'scheme': 'https'},
    ...                       {'url': 'https://en.wikipedia.org/wiki/Abinomn_language',
    ...                        'title': None, 'scheme': 'https'}],
    ...             'timespan': None,
    ...             'classification': None,
    ...             }}.items(),
    ...           dump=True, convert_lines=False))  # doctest: +NORMALIZE_WHITESPACE
    {('abin1243',):
     {'core': {'name': 'Abinomn',
               'hid': 'bsa',
               'level': 'language',
               'iso639-3': 'bsa',
               'latitude': '-2.92281',
               'longitude': '138.891',
               'macroareas': ['Papunesia'],
               'countries': ['ID'],
               'links': ['[Abinomn](http://endangeredlanguages.com/lang/1763)',
                         'https://www.wikidata.org/entity/Q56648',
                         'https://en.wikipedia.org/wiki/Abinomn_language'],
                         'timespan': None},
      'sources': {},
      'altnames': {},
      'triggers': {},
      'identifier': {},
      'classification': {},
      'endangerment': {},
      'hh_ethnologue_comment': {},
      'iso_retirement': {}}}
    >>> dict(pipe({('abin1243',): {
    ...             'id': 'abin1243',
    ...             'parent_id': None,
    ...             'name': 'Abinomn',
    ...             'level': 'language',
    ...             'hid': 'bsa',
    ...             'iso639_3': 'bsa',
    ...             'latitude': -2.92281,
    ...             'longitude': 138.891,
    ...             'macroareas': ['Papunesia'],
    ...             'countries': [{'name': 'Indonesia', 'id': 'ID'}],
    ...             'links': [{'url': 'http://endangeredlanguages.com/lang/1763',
    ...                        'title': 'Abinomn', 'scheme': 'http'},
    ...                       {'url': 'https://www.wikidata.org/entity/Q56648',
    ...                         'title': None, 'scheme': 'https'},
    ...                       {'url': 'https://en.wikipedia.org/wiki/Abinomn_language',
    ...                        'title': None, 'scheme': 'https'}],
    ...             'timespan': None,
    ...             'classification': None,
    ...             }}.items(),
    ...           dump=True, convert_lines=True))  # doctest: +NORMALIZE_WHITESPACE
    {('abin1243',):
     {'core': {'name': 'Abinomn',
               'hid': 'bsa',
               'level': 'language',
               'iso639-3': 'bsa',
               'latitude': '-2.92281',
               'longitude': '138.891',
               'macroareas': '\nPapunesia',
               'countries': '\nID',
               'links': '\n[Abinomn](http://endangeredlanguages.com/lang/1763)\nhttps://www.wikidata.org/entity/Q56648\nhttps://en.wikipedia.org/wiki/Abinomn_language',
               'timespan': None},
      'sources': {},
      'altnames': {},
      'triggers': {},
      'identifier': {},
      'classification': {},
      'endangerment': {},
      'hh_ethnologue_comment': {},
      'iso_retirement': {}}}
    """  # noqa: E501
    for path, l in languoids:
        record = make_record(l, convert_lines=convert_lines)
        yield path, record


def make_languoid(path_tuple: _globals.PathType, cfg: _globals.RecordType, /, *,
                  convert_lines: bool) -> _globals.LanguoidType:
    _make_lines = _fields.parse_lines if convert_lines else make_lines_raw

    core = cfg[CORE]

    languoid = {'id': path_tuple[-1],
                'parent_id': path_tuple[-2] if len(path_tuple) > 1 else None,
                'name': core['name'],
                'level': core['level'],
                'hid': core.get('hid'),
                'iso639_3': core.get('iso639-3'),
                'latitude': get_float(core, 'latitude'),
                'longitude': get_float(core, 'longitude'),
                'macroareas': _make_lines(core.get('macroareas')),
                'countries': sorted((splitcountry(c)
                                     for c in _make_lines(core.get('countries'))),
                                    key=operator.itemgetter('id')),
                'links': [splitlink(c) for c in _make_lines(core.get('links'))],
                'timespan': make_interval(core.get('timespan')),
                'sources': None,
                'altnames': None,
                'triggers': None,
                'identifier': None,
                'classification': None,
                'endangerment': None,
                'hh_ethnologue_comment': None,
                'iso_retirement': None}

    if SOURCES in cfg:
        sources = {provider: [splitsource(p) for p in _make_lines(sources)]
                   for provider, sources in cfg['sources'].items()}
        sources = skip_empty(sources)
        if sources:
            languoid[SOURCES] = sources

    if ALTNAMES in cfg:
        altnames = {provider: [splitaltname(a) for a in _make_lines(altnames)]
                    for provider, altnames in cfg[ALTNAMES].items()}
        if altnames:
            languoid[ALTNAMES] = altnames

    if TRIGGERS in cfg:
        triggers = {field: _make_lines(triggers)
                    for field, triggers in cfg[TRIGGERS].items()}
        if triggers:
            languoid[TRIGGERS] = triggers

    if IDENTIFIER in cfg:
        # FIXME: semicolon-separated (wals)?
        identifier = dict(cfg[IDENTIFIER])
        if identifier:
            languoid[IDENTIFIER] = identifier

    if CLASSIFICATION in cfg:
        classification = {c: (list(map(splitsource, _make_lines(classifications)))
                              if c.endswith('refs') else classifications)
                          for c, classifications in cfg[CLASSIFICATION].items()}
        classification = skip_empty(classification)
        if classification:
            languoid[CLASSIFICATION] = classification

    if ENGANGERMENT in cfg:
        sct = cfg[ENGANGERMENT]
        languoid[ENGANGERMENT] = {'status': sct['status'],
                                  'source': splitsource(sct['source'],
                                                        endangerment=True),
                                  'date': make_datetime(sct['date']),
                                  'comment': sct['comment']}

    if HH_ETHNOLOGUE_COMMENT in cfg:
        sct = cfg[HH_ETHNOLOGUE_COMMENT]
        languoid[HH_ETHNOLOGUE_COMMENT] = {'isohid': sct['isohid'],
                                           'comment_type': sct['comment_type'],
                                           'ethnologue_versions': sct['ethnologue_versions'],
                                           'comment': sct['comment']}

    if ISO_RETIREMENT in cfg:
        sct = cfg[ISO_RETIREMENT]
        languoid[ISO_RETIREMENT] = {'code': sct['code'],
                                    'name': sct['name'],
                                    'change_request': sct.get('change_request'),
                                    'effective': make_date(sct['effective']),
                                    'reason': sct['reason'],
                                    'change_to': _make_lines(sct.get('change_to')),
                                    'remedy': sct.get('remedy'),
                                    'comment': sct.get('comment')}

    return languoid


def make_record(languoid: _globals.LanguoidType, /, *,
                convert_lines: bool,
                is_lines=_fields.is_lines) -> _globals.RecordType:
    core = {'name': languoid['name'],
            'hid': languoid['hid'],
            'level': languoid['level'],
            'iso639-3': languoid['iso639_3'],
            'latitude': format_float(languoid['latitude']),
            'longitude': format_float(languoid['longitude']),
            'macroareas': languoid['macroareas'],
            'countries': list(map(formatcountry, languoid['countries'])),
            'links': list(map(formatlink, languoid['links'])),
            'timespan': format_interval(languoid.get('timespan'))}

    record = {CORE: core}

    sources = languoid.get('sources')
    if sources:
        sources = {p: list(map(formatsource, s)) for p, s in sources.items()}
    else:
        sources = {}

    altnames = languoid.get('altnames')
    if altnames:
        altnames = {p: list(map(formataltname, a)) for p, a in altnames.items()}
    else:
        altnames = {}

    triggers = languoid.get('triggers') or {}

    identifier = languoid.get('identifier') or {}

    classification = languoid['classification']
    if classification:
        classification.update({k: list(map(formatsource, classification[k]))
                               for k in ('subrefs', 'familyrefs')
                               if k in classification})
    else:
        classification = {}

    endangerment = languoid.get('endangerment')
    if endangerment:
        endangerment.update(source=formatsource(endangerment['source'],
                                                endangerment=True),
                            date=format_datetime(endangerment['date']))
    else:
        endangerment = {}

    hh_ethnologue_comment = languoid.get('hh_ethnologue_comment') or {}

    iso_retirement = languoid.get('iso_retirement')
    if iso_retirement:
        if isinstance(iso_retirement['effective'], datetime.date):
            iso_retirement['effective'] = format_date(iso_retirement['effective'])
    else:
        iso_retirement = {}

    record.update({SOURCES: sources,
                   ALTNAMES: altnames,
                   TRIGGERS: triggers,
                   IDENTIFIER: identifier,
                   CLASSIFICATION: classification,
                   ENGANGERMENT: endangerment,
                   HH_ETHNOLOGUE_COMMENT: hh_ethnologue_comment,
                   ISO_RETIREMENT: iso_retirement})

    if convert_lines:
        _, record = _fields.join_lines_inplace((None, record))

    return record


def make_lines_raw(value, /):
    """No-op.

    >>> make_lines_raw(None)
    []

    >>> make_lines_raw(['spam', 'eggs'])
    ['spam', 'eggs']
    """
    if value is None:
        return []
    return value


def skip_empty(mapping, /):
    """

    >>> skip_empty({'spam': None, 'eggs': [], 'bacon': 'ham'})
    {'bacon': 'ham'}
    """
    return {k: v for k, v in mapping.items() if v}


def get_float(mapping, key, /, *, format_=FLOAT_FORMAT):
    """

    >>> get_float({'spam': '1.42'}, 'spam')
    1.42

    >>> assert get_float({}, 'spam') is None
    >>> assert get_float({'eggs': 42}, 'spam') is None
    """
    result = mapping.get(key)
    if result is not None:
        result = float(format_ % float(result))
    return result


def format_float(value, /, *, format_=FLOAT_FORMAT):
    """

    >>> format_float(1.42)
    '1.42'

    >>> assert format_float(None) is None
    """
    if value is None:
        return None
    return str(float(format_ % value))


def make_date(value, /, *, format_=DATE_FORMAT):
    """

    >>> make_date('2001-12-31')
    datetime.date(2001, 12, 31)
    """
    return datetime.datetime.strptime(value, format_).date()


def format_date(value, /, *, format_=DATE_FORMAT):
    """

    >>> format_date(datetime.date(2001, 12, 31))
    '2001-12-31'
    """
    return value.strftime(format_)


def make_datetime(value, /, *, format_=DATETIME_FORMAT):
    """

    >>> make_datetime('2001-12-31T23:59:59')
    datetime.datetime(2001, 12, 31, 23, 59, 59)
    """
    return datetime.datetime.strptime(value, format_)


def format_datetime(value, /, *, format_=DATETIME_FORMAT):
    """

    >>> format_datetime(datetime.datetime(2001, 12, 31, 23, 59, 59))
    '2001-12-31T23:59:59'
    """
    return value.strftime(format_)


def make_interval(value, /, *, date_format=DATE_FORMAT, fix_year=True,
                  _match=ISO_8601_INTERVAL.fullmatch, strict=False):
    """

    >>> make_interval('-9999-01-01/+9999-12-31')
    {'start_year': -9999, 'start_month': 1, 'start_day': 1, 'end_year': 9999, 'end_month': 12, 'end_day': 31}

    >>> assert make_interval(None) is None
    """
    if value is None:
        return None
    value = value.strip()
    if (ma := _match(value)) is None:
        warnings.warn(f'unmatched interval: {value!r}')

        if strict:  # pragma: no cover
            log.error('invalid interval', value)
            raise ValueError('invalid interval', value)

        log.warning('ignoring interval value %r', value)
        return None

    dates = ma.group('start_date', 'end_date')
    if fix_year:
        def fix_date(d, year_tmpl='{:04d}'):
            year, sep, rest = d.partition('-')
            assert year and sep and rest
            year = year_tmpl.format(int(year))
            return f'{year}{sep}{rest}'

        dates = list(map(fix_date, dates))

    start, end = (datetime.datetime.strptime(d, date_format).date()
                  for d in dates)

    start_sign = -1 if ma.group('start_sign') == '-' else 1
    end_sign = -1 if ma.group('end_sign') == '-' else 1

    return {'start_year': start.year * start_sign,
            'start_month': start.month,
            'start_day': start.day,
            'end_year': end.year * end_sign,
            'end_month': end.month,
            'end_day': end.day}


def format_interval(value, /, *, year_tmpl='{: 05d}'):
    """

    >>> format_interval({'start_year': -9999,
    ...                  'start_month': 1,
    ...                  'start_day': 1,
    ...                  'end_year': 9999,
    ...                  'end_month': 12,
    ...                  'end_day': 31})
    '-9999-01-01/9999-12-31'

    >>> assert format_interval(None) is None
    """
    if value is None:
        return None

    year_values = [value.pop(key) for key in ('start_year', 'end_year')]
    # https://en.wikipedia.org/wiki/ISO_8601#Years
    for year in year_values:
        assert -9999 <= year <= 9999

    start_year, end_year = (year_tmpl.format(y).strip() for y in year_values)
    context = dict(value, start_year=start_year, end_year=end_year)

    return ('{start_year}-{start_month:02d}-{start_day:02d}'
            '/'
            '{end_year}-{end_month:02d}-{end_day:02d}').format_map(context)


_COUNTRY_PATTERN = re.compile(r'''
(?P<id_only>[A-Z]{2})
|
(?:
    (?P<name>.+?)
    [ ]
    \(
        (?P<id>[^)]+)
    \)
)
'''.strip(), flags=re.VERBOSE)


def splitcountry(name, /, *, _match=_COUNTRY_PATTERN.fullmatch):
    """

    >>> splitcountry('The Union of the Comoros (KM)')
    {'name': 'The Union of the Comoros', 'id': 'KM'}
    """
    groups = _match(name).groupdict()
    id_only = groups.pop('id_only')
    if id_only:
        country = pycountry.countries.get(alpha_2=id_only)
        return {'id': id_only, 'name': country.name}
    return groups


def formatcountry(value, /, *, minimal=True):
    """

    >>> formatcountry({'name': 'The Kingdom of Norway', 'id': 'NO'})
    'NO'

    >>> formatcountry({'name': 'The Kingdom of Norway', 'id': 'NO'},
    ...               minimal=False)
    'The Kingdom of Norway (NO)'
    """
    return ('{name} ({id})' if not minimal else '{id}').format_map(value)


_LINK_PATTERN = re.compile(r'''
\[
    (?P<title>[^]]+)
\]
\(
    (?P<url>[^)]+)
\)
'''.strip(), flags=re.VERBOSE)


def splitlink(markdown, /, *, _match=_LINK_PATTERN.fullmatch):
    """

    >>> splitlink('https://www.example.com')
    {'url': 'https://www.example.com', 'title': None, 'scheme': 'https'}

    >>> splitlink('http://www.example.com')
    {'url': 'http://www.example.com', 'title': None, 'scheme': 'http'}

    >>> splitlink('[Example](https://www.example.com)')
    {'url': 'https://www.example.com', 'title': 'Example', 'scheme': 'https'}

    >>> splitlink('www.example.com')
    {'url': 'www.example.com', 'title': None, 'scheme': None}
    """
    if (ma := _match(markdown)) is not None:
        title, url = ma.groups()
    else:
        title = None
        url = markdown

    scheme, sep, rest = url.partition('://')
    if sep:
        assert rest
        scheme = scheme.lower()
    else:
        scheme = None

    return {'url': url, 'title': title, 'scheme': scheme}


def formatlink(value, /):
    """

    >>> formatlink({'url': 'https://example.com'})
    'https://example.com'

    >>> formatlink({'url': 'https://example.com', 'title': 'Example'})
    '[Example](https://example.com)'
    """
    if value.get('title') is None:
        return value['url']
    return '[{title}]({url})'.format_map(value)


_SOURCE_PATTERN = re.compile(r'''
\*{2}
    (?P<bibfile>[a-z0-9_\-]+)
    :
    (?P<bibkey>[a-zA-Z0-9_\-/.;:?!'()\[\]]+?)
\*{2}
(?:
    :
    (?P<pages>
        [0-9]+(?:-[0-9]+)?
        (?:[,;][ ][0-9]+(?:-[0-9]+)?)*
    )
)?
(?:
    <trigger[ ]"
        (?P<trigger>[^\"]+)
    ">
)?
'''.strip(), flags=re.VERBOSE)


def splitsource(s, /, *, _match=_SOURCE_PATTERN.match,  # pre v4.1 compat
                endangerment=False):
    """

    >>> splitsource('**hh:42**')
    {'bibfile': 'hh', 'bibkey': '42', 'pages': None, 'trigger': None}

    >>> splitsource('**hh:42**:23-55')
    {'bibfile': 'hh', 'bibkey': '42', 'pages': '23-55', 'trigger': None}

    >>> splitsource('**hh:42**:23-55, 123-155')
    {'bibfile': 'hh', 'bibkey': '42', 'pages': '23-55, 123-155', 'trigger': None}

    >>> splitsource('**hh:42**:23-55; 123-155<trigger "spam">')
    {'bibfile': 'hh', 'bibkey': '42', 'pages': '23-55; 123-155', 'trigger': 'spam'}
    """
    if endangerment and s.isalnum():
        return {'name': s, 'bibfile': None, 'bibkey': None, 'pages': None}

    result = _match(s).groupdict()
    if endangerment:
        result['name'] = s
        result.pop('trigger', None)
    return result


def formatsource(value, /, *, endangerment=False):
    """

    >>> formatsource({'bibfile': 'hh', 'bibkey': '23'})
    '**hh:23**'

    >>> formatsource({'bibfile': 'hh', 'bibkey': '23', 'pages': '1-23'})
    '**hh:23**:1-23'

    >>> formatsource({'bibfile': 'hh', 'bibkey': '23',
    ...               'pages': '1-23', 'trigger': 'spam'})
    '**hh:23**:1-23<trigger "spam">'

    >>> formatsource({'name': 'HH'}, endangerment=True)
    'HH'
    """
    if endangerment and value.get('bibfile') is None:
        return value['name']

    result = ['**{bibfile}:{bibkey}**'.format_map(value)]
    if value.get('pages') is not None:
        result.append(':{pages}'.format_map(value))
    if value.get('trigger') is not None:
        result.append('<trigger "{trigger}">'.format_map(value))
    return ''.join(result)


_ALTNAME_PATTERN = re.compile(r'''
(?P<name>.+?)
(?:
    [ ]
     \[
         (?P<lang>[a-z]{2,3})
     \]
)?
'''.strip(), flags=re.VERBOSE)


def splitaltname(s, /, *, _match=_ALTNAME_PATTERN.fullmatch):
    """

    >>> splitaltname('Spam')
    {'name': 'Spam', 'lang': None}

    >>> splitaltname('Späm [de]')
    {'name': 'Späm', 'lang': 'de'}

    >>> splitaltname('Späm [deu]')
    {'name': 'Späm', 'lang': 'deu'}
    """
    return _match(s).groupdict()


def formataltname(value, /):
    """

    >>> formataltname({'name': 'Spam'})
    'Spam'

    >>> formataltname({'name': 'Spam', 'lang': None})
    'Spam'

    >>> formataltname({'name': 'Spam', 'lang': ''})
    'Spam'

    >>> formataltname({'name': 'Späm', 'lang': 'de'})
    'Späm [de]'

    >>> formataltname({'name': 'Späm', 'lang': 'deu'})
    'Späm [deu]'
    """
    if value.get('lang') in ('', None):
        return value['name']
    return '{name} [{lang}]'.format_map(value)



import io
import itertools

import sqlalchemy as sa

import pytest


@pytest.mark.xfail_glottolog_tag('v4.1', reason='no timespan in Glottolog v4.1',
                                 raises=sa.exc.NoResultFound)
def test_select_languoid_timespan(treedb):
    timespan = treedb.queries.select_languoid_timespan(as_json=False,
                                                       label='ts')
    select = (sa.select(treedb.Languoid, timespan)
              .where(timespan != sa.null()).limit(1))
    with treedb.connect() as conn:
        row = conn.execute(select).one()

    assert row.ts


@pytest.mark.parametrize(
    'kwargs, expected_head',
    [pytest.param({'parent_level': 'top', 'child_level': 'language'},
                  [('abin1243', []),
                   ('abis1238', []),
                   ('abkh1242', ['abaz1241', 'abkh1244',
                                             'adyg1241', 'kaba1278',
                                             'ubyk1235']),
                   ('adai1235', [])],
                  id='parent_level=top, child_level=language')])
def test_iterdescendants(treedb, kwargs, expected_head):
    pairs = treedb.iterdescendants(**kwargs)
    head = list(itertools.islice(pairs, len(expected_head)))

    assert head == expected_head


@pytest.mark.parametrize(
    'as_rows, sort_keys',
    [(True, False),
     (True, True),
     (False, False),
     (False, True)])
def test_languoid_query_print_rows_pretty(treedb, as_rows, sort_keys):
    query = (treedb.get_languoids_query(as_rows=as_rows, sort_keys=sort_keys)
             .where(treedb.Languoid.id == 'abin1243'))
    with io.StringIO() as f:
        treedb.print_rows(query, pretty=True, file=f)
        result = f.getvalue()
    assert result
    if as_rows:
        if sort_keys:
            assert result.startswith("{'__path__': 'abin1243',\n"
                                     " 'languoid': {'altnames': {")
        else:
            assert result.startswith("{'__path__': 'abin1243',\n"
                                     " 'languoid': {'id': 'abin1243',\n"
                                     "              'parent_id': None,\n")
    else:
        if sort_keys:
            assert result.startswith("{'md.ini': {'__path__': ['abin1243'],\n"
                                     "            'languoid': {'altnames': {")
        else:
            assert result.startswith("{'md.ini': {'__path__': ['abin1243'],\n"
                                     "            'languoid': {'id': 'abin1243',\n"
                                     "                         'parent_id': None,\n")



import collections
import functools

import pytest

ZeroDict = functools.partial(collections.defaultdict, int)

FILES_WRITTEN = {'files': ZeroDict(),
                 'raw_lines': ZeroDict(),
                 'raw': ZeroDict(),
                 'tables': ZeroDict({  # replace 'Country Name (ID)' with 'ID'
                                     'v4.3-treedb-fixes': 8_528,
                                     'v4.2.1': 8_528,
                                     'v4.2': None,
                                     'v4.1': 8_540})}


pytestmark = pytest.mark.writes


@pytest.mark.parametrize(
    'source',
    ['files',
     pytest.param('raw_lines', marks=pytest.mark.raw),
     pytest.param('raw', marks=pytest.mark.raw),
     'tables'],
    ids=lambda x: f'source={x}')
def test_write_files(pytestconfig, treedb, source):
    expected = FILES_WRITTEN[source].get(pytestconfig.option.glottolog_tag)

    files_written = treedb.write_files(source=source,
                                       dry_run=True,
                                       require_nwritten=expected,
                                       bind=treedb.engine)
    if expected is not None:
        assert files_written == expected
    else:
        assert 0 <= files_written <= 40_000



"""Generic re-useable self-contained helpers."""

from collections.abc import Iterator
import builtins
import bz2
import configparser
import contextlib
import functools
import gzip
import hashlib
import io
import itertools
import json
import logging
import lzma
import operator
import os
import pathlib
import platform
import subprocess
import sys
import types  # noqa: F401 (used in doctest)
import warnings

ENCODING = 'utf-8'

PROGRESS_AFTER = 5_000

SUFFIX_OPEN_MODULE = {'.bz2': bz2,
                      '.gz': gzip,
                      '.xz': lzma}

__all__ = ['uniqued',
           'next_count',
           'groupby_itemgetter', 'groupby_attrgetter',
           'islice_limit',
           'iterslices',
           'walk_scandir',
           'pipe_json_lines', 'pipe_json', 'pipe_lines',
           'get_open_module',
           'path_from_filename',
           'sha256sum',
           'run',
           'Ordering',
           'ConfigParser']


log = logging.getLogger(__name__)


def uniqued(iterable, /):
    """Return list of unique hashable elements preserving order.

    >>> uniqued('spamham')
    ['s', 'p', 'a', 'm', 'h']
    """
    seen = set()
    return [i for i in iterable if i not in seen and not seen.add(i)]


def next_count(*, start: int = 0, step: int = 1):
    """Return a callable returning descending ints.

    >>> nxt = next_count(start=1)

    >>> nxt()
    1

    >>> nxt()
    2
    """
    count = itertools.count(start, step)
    return functools.partial(next, count)


def groupby_itemgetter(*indexes):
    """

    >>> groupby_second = groupby_itemgetter(1)

    >>> people = [('Sir Robin', True),
    ...           ('Brian', False),
    ...           ('Sir Lancelot', True)]

    >>> {knight: [name for name, _ in grp] for knight, grp in groupby_second(people)}
    {True: ['Sir Lancelot'], False: ['Brian']}

    >>> people_sorted = sorted(people, key=operator.itemgetter(1))

    >>> {knight: [name for name, _ in grp] for knight, grp in groupby_second(people_sorted)}
    {False: ['Brian'], True: ['Sir Robin', 'Sir Lancelot']}
    """
    key = operator.itemgetter(*indexes)
    return functools.partial(itertools.groupby, key=key)


def groupby_attrgetter(*attrnames):
    """

    >>> groupby_knight = groupby_attrgetter('knight')

    >>> people = [types.SimpleNamespace(name='Sir Robin', knight=True),
    ...           types.SimpleNamespace(name='Brian', knight=False),
    ...           types.SimpleNamespace(name='Sir Lancelot', knight=True)]

    >>> {knight: [g.name for g in grp] for knight, grp in groupby_knight(people)}
    {True: ['Sir Lancelot'], False: ['Brian']}

    >>> people_sorted = sorted(people, key=operator.attrgetter('knight'))

    >>> {knight: [g.name for g in grp] for knight, grp in groupby_knight(people_sorted)}
    {False: ['Brian'], True: ['Sir Robin', 'Sir Lancelot']}
    """
    key = operator.attrgetter(*attrnames)
    return functools.partial(itertools.groupby, key=key)


def islice_limit(iterable, /, *,
                 limit: int | None = None,
                 offset: int | None = 0):
    """Return a slice from iterable applying limit and offset.

    >>> list(islice_limit('spam', limit=3))
    ['s', 'p', 'a']

    >>> list(islice_limit('spam', offset=3))
    ['m']

    >>> list(islice_limit('spam', offset=1, limit=2))
    ['p', 'a']

    >>> list(islice_limit('spam'))
    ['s', 'p', 'a', 'm']
    """
    if limit is not None and offset:
        stop = offset + limit
        return itertools.islice(iterable, offset, stop)
    elif limit is not None:
        return itertools.islice(iterable, limit)
    elif offset:
        return itertools.islice(iterable, offset, None)
    return iterable


def iterslices(iterable, /, *, size: int):
    """Yield iterable in chunks of maximal size.

    >>> [tuple(chunk) for chunk in iterslices('bacon', size=2)]
    [('b', 'a'), ('c', 'o'), ('n',)]
    """
    iterable = iter(iterable)
    next_slice = functools.partial(itertools.islice, iterable, size)
    return iter(lambda: list(next_slice()), [])


def walk_scandir(top, /, *,
                 verbose: bool = False,
                 sortkey=operator.attrgetter('name')) -> Iterator[os.DirEntry]:
    """Yield os.DirEntry objects for all files under top."""
    # NOTE: os.walk() ignores errors and this can be more efficient
    top = path_from_filename(top)
    if not top.is_absolute():
        top = pathlib.Path.cwd().joinpath(top).resolve()

    if platform.system() == 'Windows':
        top = pathlib.Path(fr'\\?\{top}')

    log.debug('recursive depth-first scandir on %r', top)
    log.debug('sortkey: %r', sortkey)

    stack = [str(top)]

    while stack:
        root = stack.pop()
        if verbose:  # pragma: no cover
            print(root)

        with os.scandir(root) as dentries:
            dentries = sorted(dentries, key=sortkey)

        dirs = []
        for dentry in dentries:
            if dentry.is_dir():
                dirs.append(dentry.path)
            else:
                yield dentry

        stack.extend(reversed(dirs))


def pipe_json_lines(file, documents=None, /, *,
                    delete_present: bool = True,
                    autocompress: bool = True,
                    newline: str | None = '\n',
                    sort_keys: bool = True,
                    compact: bool = True,
                    indent: int | None = None,
                    ensure_ascii: bool = False):
    r"""Load/dump json lines as endpoint pipe.

    >>> with io.StringIO() as f:
    ...     buf, nwritten = pipe_json_lines(f, [None, {'spam': None}])
    ...     json_lines = f.getvalue()

    >>> assert buf is f

    >>> nwritten
    2

    >>> json_lines
    'null\n{"spam":null}\n'

    >>> with io.StringIO(json_lines) as f:
    ...     documents = list(pipe_json_lines(f))
    >>> documents
    [None, {'spam': None}]
    """
    lines_kwargs = {'delete_present': delete_present,
                    'autocompress': autocompress,
                    'newline': newline}
    json_kwargs = {'sort_keys': sort_keys,
                   'ensure_ascii': ensure_ascii,
                   'compact': compact,
                   'indent': indent}

    if documents is not None:
        lines = pipe_json(documents, dump=True, **json_kwargs)
        return pipe_lines(file, lines, **lines_kwargs)

    lines = pipe_lines(file, **lines_kwargs)
    return pipe_json(lines, dump=False, **json_kwargs)


def pipe_json(documents, /, *, dump: bool,
              sort_keys: bool = True,
              compact: bool = False,
              indent: int | None = None,
              ensure_ascii: bool = False):
    """Bidirectional codec between a generator and a consumer."""
    codec = json.dumps if dump else json.loads

    if dump:
        dump_kwargs = {'sort_keys': sort_keys,
                       'indent': indent,
                       'ensure_ascii': ensure_ascii,
                       # json-serialize datetime.datetime
                       'default': operator.methodcaller('isoformat')}
        if compact:
            if indent:  # pragma: no cover
                warnings.warn(f'{indent=!r} overridden by {compact=}')
            dump_kwargs.update(indent=None,
                               separators=(',', ':'))
        elif indent:
            warnings.warn('non-canonical JSON Lines format from indent %r' % indent)

        codec = functools.partial(codec, **dump_kwargs)

    def itercodec(docs):
        return map(codec, docs)

    if dump:
        assert next(itercodec([None])) == 'null'
    else:
        assert next(itercodec(['null'])) is None

    return itercodec(documents)


def pipe_lines(file, lines=None, /, *, newline: str | None = None,
               delete_present: bool = False, autocompress: bool = True):
    open_func, result, hashobj = get_open_result(file,
                                                 write=lines is not None,
                                                 delete_present=delete_present,
                                                 autocompress=autocompress,
                                                 newline=newline)

    if lines is not None:
        with open_func() as f:
            if hashobj is not None:
                total = write_wrapped(hashobj, f, lines)
            else:
                total = write_lines(f, lines)

            if file is None:
                result = f.getvalue()
        return result, total

    def iterlines():
        with open_func() as f:
            yield from f

    return iterlines()


def write_wrapped(hashsum, f, lines, /, *, buflines: int = 1_000) -> int:
    write_line = functools.partial(print, file=f)
    buf = f.buffer
    total = 0
    for lines in iterslices(lines, size=buflines):
        for line in lines:
            write_line(line)
        total += len(lines)
        hashsum.update(buf.getbuffer())
        # NOTE: f.truncate(0) would prepend zero-bytes
        f.seek(0)
        f.truncate()
    return total


def write_lines(file, lines, /) -> int:
    r"""

    >>> with io.StringIO() as f:
    ...    write_lines(f, ['spam', 'eggs'])
    ...    text = f.getvalue()
    2

    >>> text
    'spam\neggs\n'
    """
    write_line = functools.partial(print, file=file)
    total = 0
    for total, line in enumerate(lines, start=1):
        write_line(line)
    return total


def path_from_filename(filename, /, *args, expanduser: bool = True):
    if hasattr(filename, 'open'):
        assert not args
        result = filename
    else:
        result = pathlib.Path(filename, *args)

    if expanduser:
        result = result.expanduser()
    return result


def get_open_result(file, /, *, write: bool = False,
                    delete_present: bool = False,
                    autocompress: bool = False,
                    newline: str | None = None,
                    _encoding: str = 'utf-8'):
    open_kwargs = {'mode': 'wt' if write else 'rt',
                   'encoding': _encoding,
                   'newline': newline}
    textio_kwargs = {'write_through': True, 'encoding': _encoding}

    path = fobj = hashobj = None

    if file is None:
        if not write:  # pragma: no cover
            raise TypeError('file cannot be Null for write=False')
        result = fobj = io.StringIO()
    elif file is sys.stdout:
        result = fobj = io.TextIOWrapper(sys.stdout.buffer,
                                         newline=open_kwargs.pop('newline', None),
                                         **textio_kwargs)
    elif hasattr(file, 'write'):
        result = fobj = file
    elif hasattr(file, 'hexdigest'):
        if not write:  # pragma: no cover
            raise TypeError('missing lines')
        result = hashobj = file
        fobj = io.TextIOWrapper(io.BytesIO(),
                                newline=open_kwargs.pop('newline', None),
                                **textio_kwargs)
    else:
        result = path = path_from_filename(file)

    if path is None:
        log.info('write lines into: %r', fobj)

        def open_func():
            return contextlib.nullcontext(fobj)
    else:
        log.info('write lines: %r', path)
        open_module = get_open_module(path, autocompress=autocompress)
        open_func = functools.partial(open_module.open, path,
                                      **open_kwargs)

        if write and path.exists():
            if not delete_present:  # pragma: no cover
                raise RuntimeError('refuse to delete_present file: {path!r}')
            warnings.warn(f'delete present file: {path!r}')
            path.unlink()

    assert result is not None

    return open_func, result, hashobj


def get_open_module(filepath, /, *, autocompress: bool = False):
    file = path_from_filename(filepath)

    suffix = file.suffix.lower()
    if autocompress:
        result = SUFFIX_OPEN_MODULE.get(suffix, builtins)
    else:
        result = builtins
        if suffix in SUFFIX_OPEN_MODULE:  # pragma: no cover
            warnings.warn(f'file {file!r} has suffix {suffix!r}'
                          ' but autocompress=False')
    return result


def sha256sum(file, /, *, raw: bool = False,
              autocompress: bool = True,
              hash_file_string: bool = False,
              file_string_encoding: str = ENCODING):
    """

    >>> sha256sum('', hash_file_string=True)
    'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
    """
    hashobj = hashlib.sha256()

    if hash_file_string:
        encoded_bytes = file.encode(file_string_encoding)
        hashobj.update(encoded_bytes)
    else:
        file = path_from_filename(file)
        open_module = get_open_module(file, autocompress=autocompress)

        with open_module.open(file, 'rb') as f:
            update_hashobj(hashobj, f)

    return hashobj if raw else hashobj.hexdigest()


def update_hashobj(hashobj, file, /, *, chunksize: int = 2**16):  # 64 KiB
    for chunk in iter(functools.partial(file.read, chunksize), b''):
        hashobj.update(chunk)


def run(cmd, /, *, capture_output: bool = False,
        unpack: bool = False,
        cwd=None,
        check: bool = False,
        encoding: str = ENCODING):
    log.info('subprocess.run(%r)', cmd)

    if platform.system() == 'Windows':
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        startupinfo.wShowWindow = subprocess.SW_HIDE
    else:
        startupinfo = None

    kwargs = {'cwd': cwd, 'encoding': encoding, 'startupinfo': startupinfo}

    if capture_output:
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, **kwargs)
        out, err = proc.communicate()
        if check and proc.returncode:
            raise subprocess.CalledProcessError(proc.returncode, cmd,
                                                output=out, stderr=err)
        proc = subprocess.CompletedProcess(cmd, proc.returncode, out, err)
        if unpack:
            return proc.stdout.strip()
    else:
        proc = subprocess.run(cmd, check=check, **kwargs)
    return proc


class Ordering(dict):
    """Dict returning infinity for unknown keys ordering them lexicographically.

    >>> o = Ordering.fromlist(['spam', 'eggs', 'bacon'])

    >>> o
    {'spam': 0, 'eggs': 1, 'bacon': 2}

    >>> o['ham']
    inf

    >>> o.sorted(['ham', 'bacon', 'eggs', 'am'])
    ['eggs', 'bacon', 'am', 'ham']

    >>> list(o.sorted_enumerate(['ham', 'bacon', 'eggs'], start=1))
    [(1, 'eggs'), (2, 'bacon'), (3, 'ham')]
    """

    _missing = float('inf')

    @classmethod
    def fromlist(cls, keys, /, *, start_index: int = 0):
        return cls((k, i) for i, k in enumerate(uniqued(keys), start=start_index))

    def __missing__(self, key, /):
        return self._missing

    def _sortkey(self, key, /):
        return self[key], key

    def sorted(self, keys, /):
        return sorted(keys, key=self._sortkey)

    def sorted_enumerate(self, keys, /, *, start: int = 0):
        keyed = sorted((self[key], key) for key in keys)
        return ((i, key) for i, (_, key) in enumerate(keyed, start=start))


class ConfigParser(configparser.ConfigParser):
    """Conservative ConfigParser with optional encoding header."""

    _basename = None

    _init_defaults = {'delimiters': ('=',),
                      'comment_prefixes': ('#',),
                      'interpolation': None}

    _newline = None

    _header = None

    @classmethod
    def from_file(cls, filename, /, *, encoding=ENCODING, **kwargs):
        path = path_from_filename(filename)
        if cls._basename is not None and path.name != cls._basename:
            raise RuntimeError(f'unexpected filename {path!r}'
                               f' (must end with {cls._basename})')

        inst = cls(**kwargs)
        with path.open(encoding=encoding) as f:
            inst.read_file(f)
        return inst

    def __init__(self, /, *, defaults=None, **kwargs):
        for k, v in self._init_defaults.items():
            kwargs.setdefault(k, v)
        super().__init__(defaults=defaults, **kwargs)

    def to_dict(self, /, *, sort_sections: bool = False,
                _default_section: str = configparser.DEFAULTSECT):
        items = sorted(self.items()) if sort_sections else self.items()
        return {name: dict(section) for name, section in items
                if name != _default_section}

    def to_file(self, filename, /, *, encoding=ENCODING):
        path = path_from_filename(filename)
        with path.open('wt', encoding=encoding, newline=self._newline) as f:
            if self._header is not None:
                f.write(self._header.format(encoding=encoding))
            self.write(f)



"""Limited ``sqlalchemy`` support for views based on https://github.com/sqlalchemy/sqlalchemy/wiki/Views."""

import functools
import logging
import types

import sqlalchemy as sa
import sqlalchemy.ext.compiler

from .. import _globals

__all__ = ['register_view',
           'create_all_views',
           'view',
           'make_table']

REGISTERED = {}

DDL = {}

TABLES = types.SimpleNamespace()


log = logging.getLogger(__name__)


def register_view(name, **kwargs):
    log.debug('register_view(%r)', name)
    assert name not in REGISTERED

    def decorator(func):
        REGISTERED[name] = functools.partial(func, **kwargs)
        return func

    return decorator


def create_all_views(*, clear: bool = False):
    log.debug('run create_view() for %d views in REGISTERED', len(REGISTERED))
    for name, func in REGISTERED.items():
        table = view(name, selectable=func(), clear=clear)
        setattr(TABLES, name, table)


def view(name, selectable, *, clear: bool = False):
    """Register a CREATE and DROP VIEW DDL for the given selectable."""
    log.debug('view(%r, clear=%r)', name, clear)

    if clear:
        DDL[name] = None, None
        return None

    DDL[name] = (CreateView(name, selectable),
                 DropView(name))

    return make_table(selectable, name=name)


def make_table(selectable, /, *, name: str = 'view_table'):
    table = sa.table(name)
    for c in selectable.alias().c:
        _, col = c._make_proxy(table,
                               primary_key=table.primary_key,
                               foreign_keys=table.foreign_keys)
        table.append_column(col)
    return table


@sa.event.listens_for(_globals.REGISTRY.metadata, 'after_create')
def after_create(target, bind, **kwargs):
    for name, (create_view, _) in DDL.items():
        if create_view is not None:
            log.debug('CREATE VIEW %r', name)
            create_view(target, bind)


@sa.event.listens_for(_globals.REGISTRY.metadata, 'before_drop')
def before_drop(target, bind, **kwargs):
    for name, (_, drop_view) in DDL.items():
        if drop_view is not None:
            log.debug('DROP VIEW %r', name)
            drop_view(target, bind)


class CreateView(sa.schema.DDLElement):

    def __init__(self, name, selectable):
        self.name = name
        self.selectable = selectable


class DropView(sa.schema.DDLElement):

    def __init__(self, name):
        self.name = name


@sa.ext.compiler.compiles(CreateView)
def compile_create_view(element, compiler, **kwargs):
    select = compiler.sql_compiler.process(element.selectable,
                                           literal_binds=True)
    return f'\nCREATE VIEW {element.name} AS {select}\n'


@sa.ext.compiler.compiles(DropView)
def compile_drop_view(element, compiler, **kwargs):
    return f'\nDROP VIEW {element.name}\n'



#!/usr/bin/env python3
# flake8: noqa

"""Some ``treedb`` usage examples."""

import pathlib
import pprint

import sqlalchemy as sa
import treedb

DBFILE = None  # pathlib.Path('treedb.sqlite3')


#treedb.configure_logging(level='INFO')
treedb.configure(log_sql=False)

treedb.set_root('../glottolog')
pprint.pprint(dict(treedb.iterlanguoids(limit=1)))

engine = treedb.load(DBFILE)
assert treedb.engine is engine

treedb.check()  # run sanity checks

treedb.print_rows(sa.select(treedb.Languoid).order_by('id').limit(5))

query = treedb.get_example_query()  # big example query containing 'everything'

qf = treedb.pd_read_sql(query, index_col='id')
if qf is not None:
    qf.info(memory_usage='deep')

lf = treedb.pd_read_languoids()
if lf is not None:
    lf.info(memory_usage='deep')

#treedb.write_csv()

#treedb.write_languoids()

#treedb.write_files()

import treedb.raw
pprint.pprint(dict([next(treedb.raw.fetch_records())]))

#treedb.raw.write_files()

#treedb.languoids.files.roundtrip()

treedb.print_rows(treedb.get_languoids_query(limit=1), pretty=True)



import textwrap

import pytest
import sqlalchemy as sa

from helpers import assert_file_size_between

QUERY_HASH = {'master': None,
              'v5.3': '4c9711d7da62c565250bb473bfd0c493d752e37e04c00d61dcc06db625e7fdcf',
              'v5.2.1': 'ced50816870c5da047d4533187c30224786da02a2b7372140545a5a73cb98012',
              'v5.2': 'ac5712e082e4bf4c4b341d232a4309c5a968b9f93d1c23b9cd4a1ae9faed5197',
              'v5.1': 'c235e4b709b009a31f5a8eace0edeff02b08546f9f81b90139d017723fb4fa65',
              'v5.0': 'f20f1bd9e397f614b24276c189f072fbe28be8c62b6d5f3011546735aecb5bc6',
              'v4.8': '705b1e4e30c35c659e536f0b41dc2c4197c72037b2c8eec7af20c3bf9ef4c992',
              'v4.7': 'e1054be01c4e17150e875d1cab40c754c0e1aeb3b5343a36c5771d3587115537',
              'v4.6': '4479476b397fa7dfbfd560a4bef5be06513ddc54f7c103d1f565e3a26404a90f',
              'v4.5': 'b36b5cdc7508b8c2843a3c93ff536ead9b4b63cdf80679f70993a5dd524a8926',
              'v4.4': '224691678e1f2e18406d6dd1a278e062c683ac12ec2acf57d501931d3661142e',
              'v4.3-treedb-fixes': 'bf8af9e4840642f4622cec41bf3156afac75317740ff0eef1ac75ec1998d4f78',
              'v4.2.1': '25222b4feb2d89b4edaeecc546280a059ae6ba69da961d56ee4a387ba3b81fc0',
              'v4.1': '55e9cab42b012048ae9f6c08353752fded7054bf531039979c6803ede54b95ac'}


def test_print_dataset(capsys, treedb):
    assert treedb.print_dataset() is None

    out, err = capsys.readouterr()
    assert not err

    assert out.startswith("git describe '")


@pytest.mark.slow
def test_backup(treedb):
    path = treedb.engine.file_with_suffix('.backup.sqlite3')

    engine = treedb.backup(path.name)

    assert engine.url.database == path.name
    assert_file_size_between(path, 10, 200)

    # SQLiteEngineProxy
    engine = treedb.engine.__class__(engine, future=treedb.engine.future)

    assert engine.file_exists()

    assert engine.file_mtime()

    assert engine.file_size() == path.stat().st_size
    assert 10 <= engine.file_size(as_megabytes=True) <= 200

    assert len(engine.file_sha256()) == 64


@pytest.mark.slow
def test_dump_sql(pytestconfig, treedb):
    suffix = '-memory' if treedb.engine.file is None else ''
    suffix += pytestconfig.option.file_engine_tag

    path = treedb.dump_sql()

    assert path.name == f'treedb{suffix}.sql.gz'
    assert_file_size_between(path, 1, 20)


@pytest.mark.slow
def test_csv_zipfile(pytestconfig, treedb):
    suffix = '-memory' if treedb.engine.file is None else ''
    suffix += pytestconfig.option.file_engine_tag

    path = treedb.csv_zipfile()

    assert path.name == f'treedb{suffix}.zip'
    assert_file_size_between(path, 1, 20)


def test_print_rows(capsys, treedb):
    query = (sa.select(treedb.Languoid)
             .where(treedb.Languoid.iso639_3 == 'bsa'))

    format_ = '{id}: {name} ({level}) [{iso639_3}]'

    assert treedb.print_rows(query, format_=format_, verbose=True) is None

    out, err = capsys.readouterr()
    assert not err

    expected = textwrap.dedent(
        '''
        SELECT languoid.id, languoid.name, languoid.level, languoid.parent_id, languoid.hid, languoid.iso639_3, languoid.latitude, languoid.longitude 
        FROM languoid 
        WHERE languoid.iso639_3 = :iso639_3_1
        abin1243: Abinomn (language) [bsa]
        ''').lstrip()  # noqa: E501,W291
    assert out == expected


def test_write_csv(pytestconfig, treedb):
    expected = QUERY_HASH.get(pytestconfig.option.glottolog_tag)
    suffix = '-memory' if treedb.engine.file is None else ''
    suffix += pytestconfig.option.file_engine_tag

    path = treedb.write_csv()

    assert path.name == f'treedb{suffix}.query.csv'
    assert_file_size_between(path, 1, 30)

    if expected is None:
        pass
    else:
        shasum = treedb.sha256sum(path)
        assert shasum == expected


def test_hash_csv(pytestconfig, treedb):
    expected = QUERY_HASH.get(pytestconfig.option.glottolog_tag)

    result = treedb.hash_csv()

    if expected is None:
        assert len(result) == 64
    else:
        assert result == expected



import io
import re
import textwrap

import pytest
import sqlalchemy as sa

import treedb as _treedb


def test_print_schema(capsys, bare_treedb):
    assert bare_treedb.print_schema() is None

    out, err = capsys.readouterr()
    assert not err

    assert out.strip().startswith('CREATE TABLE __dataset__ (')


@pytest.mark.parametrize(
    'query, pretty, expected',
    [pytest.param(sa.select(_treedb.Languoid), False,
                  ('SELECT languoid.id,'
                   ' languoid.name,'
                   ' languoid.level,'
                   ' languoid.parent_id,'
                   ' languoid.hid,'
                   ' languoid.iso639_3,'
                   ' languoid.latitude,'
                   ' languoid.longitude \n'
                   'FROM languoid\n'),
                  id='query=Languoid'),
     pytest.param(sa.select(_treedb.Languoid), True,
                  ('SELECT languoid.id,\n'
                   '       languoid.name,\n'
                   '       languoid.level,\n'
                   '       languoid.parent_id,\n'
                   '       languoid.hid,\n'
                   '       languoid.iso639_3,\n'
                   '       languoid.latitude,\n'
                   '       languoid.longitude\n'
                   'FROM languoid\n'),
                  id='query=Languoid, pretty', marks=pytest.mark.sqlparse),
     pytest.param(None, False, None, id='query=None'),
     pytest.param(None, True, None, id='query=None, pretty')])
def test_print_query_sql(capsys, bare_treedb, query, pretty, expected):
    assert bare_treedb.print_query_sql(query, pretty=pretty) is None

    out, err = capsys.readouterr()
    assert not err

    if expected is None:
        assert out.strip()
    elif pretty and _treedb.backend.sqlparse is None:
        norm_out, norm_exp = (re.sub(r'\s{1,}', r' ', s).strip()
                              for s in (out, expected))
        assert norm_out == norm_exp
    else:
        assert out == expected


def test_print_rows_pretty(bare_treedb, encoding='utf-8'):
    value = dict.fromkeys(f'key_{i}' for i in range(10))
    value = _treedb.backend.json_object(sort_keys_=False,
                                        load_json_=True,
                                        label_='record',
                                        **value)
    query = sa.select(value)
    with io.StringIO() as f:
        bare_treedb.print_rows(query, pretty=True, file=f)
        result = f.getvalue()

    expected = textwrap.dedent(
        '''
        {'record': {'key_0': None,
                    'key_1': None,
                    'key_2': None,
                    'key_3': None,
                    'key_4': None,
                    'key_5': None,
                    'key_6': None,
                    'key_7': None,
                    'key_8': None,
                    'key_9': None}}
        ''').lstrip()
    assert result == expected



"""Load configuration from ``treedb.ini``."""

import configparser
import logging
import os

from . import _globals
from . import _tools

__all__ = ['configure',
           'get_default_root']

ROOT_OPTION = ('glottolog', 'repo_root')

ENGINE_OPTION = ('treedb', 'engine')

NOT_SET = object()


log = logging.getLogger(__name__)


class ConfigParser(configparser.ConfigParser):

    @classmethod
    def from_file(cls, path, /, *, required=False, default_repo_root):
        path = _tools.path_from_filename(path).resolve()

        defaults = {'here': path.parent.as_posix()}
        inst = cls(defaults=defaults)
        inst.set_default_repo_root(default_repo_root)

        found = inst.read([path])
        if found:
            log.debug('%s().read() config file(s): %r', cls.__name__, found)
        elif required:
            raise ValueError(f'config file not found: {path!r}')
        else:
            log.debug('no config file(s) found')
        return inst

    def set_default_repo_root(self, repo_root):
        self.add_section(ROOT_OPTION[0])
        self.set(*ROOT_OPTION, repo_root)


def configure(config_path=_globals.CONFIG, /, *,
              engine=NOT_SET, root=NOT_SET,
              loglevel=None, log_sql: bool = None,
              default_repo_root=_globals.DEFAULT_ROOT,
              title: str | None = None,
              title_memory_tag: str = _globals.MEMORY_TAG) -> None:
    """Set root, and engine and configure logging from the given .ini file."""
    log.info('configure from %r, title=%r', config_path, title)
    log.debug('default repo root: %r', default_repo_root)

    from . import backend
    from . import languoids
    from . import logging_

    config_path = _tools.path_from_filename(config_path)
    log.debug('load config file %r', config_path)
    cfg = ConfigParser.from_file(config_path,
                                 required=True,
                                 default_repo_root=default_repo_root)

    log.info('configure logging from %r', config_path)
    logging_.configure_logging_from_file(cfg, level=loglevel, log_sql=log_sql)

    if engine is NOT_SET:
        engine = cfg.get(*ENGINE_OPTION, fallback=None)

    if engine is not None:
        engine = _tools.path_from_filename(engine)
        if not engine.is_absolute():
            engine = config_path.parent / engine

    backend.set_engine(engine, title=title, title_memory_tag=title_memory_tag)

    if root is NOT_SET:
        root = cfg.get(*ROOT_OPTION)
    root = _tools.path_from_filename(root)

    if not root.is_absolute():
        root = config_path.parent / root

    languoids.set_root(root)


def get_default_root(*, env_var,
                     config_path=_globals.CONFIG,
                     fallback=_globals.DEFAULT_ROOT):
    """Return default root from environment variable, config, or fallback."""
    root = os.getenv(env_var)

    if root is None:
        log.debug('get %r from optional config file %r', ROOT_OPTION, config_path)
        cfg = ConfigParser.from_file(config_path, default_repo_root=fallback)
        root = cfg.get(*ROOT_OPTION)

    return root



"""Stdlib logging configuration short-cuts."""

import logging
import logging.config

__all__ = ['configure_logging',
           'configure_logging_from_file']

WARNINGS = 'py.warnings'

SQL = 'sqlalchemy.engine'

FORMAT = '[%(levelname)s@%(name)s] %(message)s'


log = logging.getLogger(__name__)


def configure_logging_from_file(path, /, *,
                                level=None,
                                log_sql=None,
                                capture_warnings: bool = True,
                                reset: bool = True):
    if reset:
        reset_logging()

    log.debug('logging.config.fileConfig(%r)', path)
    logging.config.fileConfig(path, disable_existing_loggers=reset)

    package_logger = logging.getLogger(__package__)

    if level is not None:  # pragma: no cover
        if not package_logger.handlers:
            log.debug('set level of %r to %r', package_logger, level)
            package_logger.setLevel(level)
        else:
            for h in package_logger.handlers:
                if not isinstance(h, logging.FileHandler):
                    log.debug('set level of %r to %r', h, level)
                    h.setLevel(level)

    if log_sql is not None:  # pragma: no cover
        sql_logger = logging.getLogger(SQL)
        log.debug('set level of %r to %r', sql_logger, level)
        sql_logger.setLevel('INFO' if log_sql else 'WARNING')

    log_version()

    set_capture_warnings(capture_warnings)

    return package_logger


def reset_logging():
    root_logger = logging.getLogger()
    for h in list(root_logger.handlers):
        log.debug('%r.removeHandler(%r)', root_logger, h)
        root_logger.removeHandler(h)
        h.close()


def log_version(*, also_print: bool = False, print_file=None):
    import treedb

    log.info('%s version: %s', __package__, treedb.__version__)
    log.debug('%r', treedb)
    if also_print or print_file is not None:
        print(f'{__package__} version: {treedb.__version__}',
              file=print_file)


def set_capture_warnings(value: bool = True, /):
    log.debug('set logging.captureWarnings(%r)', value)
    logging.captureWarnings(value)


def configure_logging(*,
                      level='WARNING',
                      log_sql=False,
                      format=FORMAT,
                      capture_warnings=True,
                      reset=True):  # pragma: no cover
    cfg = {'version': 1,
           'root': {'handlers': ['stderr_pretty'], 'level': level},
           'loggers': {WARNINGS: {},
                       SQL: {'handlers': ['stderr_plain'],
                             'level': 'INFO' if log_sql else 'WARNING',
                             'propagate': False},
                       __package__: {}},
           'handlers': {'stderr_plain': {'formatter': 'plain',
                                         'class': 'logging.StreamHandler'},
                        'stderr_pretty': {'formatter': 'pretty',
                                          'class': 'logging.StreamHandler'}},
           'formatters': {'plain': {'format': '%(message)s'},
                          'pretty': {'format': format}}}

    if reset:
        reset_logging()

    log.debug('logging.config.dictConfig(%r)', cfg)
    logging.config.dictConfig(cfg)

    log_version()

    set_capture_warnings(capture_warnings)

    return logging.getLogger(__package__)



import pytest


def test_check(treedb):
    assert treedb.check()


@pytest.mark.skip('TODO: improve compare_languoids() fail output')
@pytest.mark.raw
def test_compare_languoids(treedb):
    assert treedb.compare_languoids()



import re

import pytest
import sqlalchemy as sa

import treedb.models as _models

BOOK = 'book1242'

WALI = 'wali1263'

TREE = {BOOK: [(BOOK, BOOK, 0, True)],
        WALI: [(WALI, WALI, 0, False),
               (WALI, 'sout3209', 1, False),
               (WALI, 'cent2291', 2, False),
               (WALI, 'daga1276', 3, False),
               (WALI, 'safa1246', 4, False),
               (WALI, 'nort3234', 5, False),
               (WALI, 'nucl1748', 6, False),
               (WALI, 'west2461', 7, False),
               (WALI, 'gurm1247', 8, False),
               (WALI, 'nucl1743', 9, False),
               (WALI, 'otiv1239', 10, False),
               (WALI, 'bwam1248', 11, False),
               (WALI, 'nort2777', 12, False),
               (WALI, 'cent2243', 13, False),
               (WALI, 'gura1261', 14, False),
               (WALI, 'nort3149', 15, False),
               (WALI, 'volt1241', 16, False),
               (WALI, 'atla1278', 17, True)]}

FULL = {'include_self': True, 'with_steps': True, 'with_terminal': True}

EXCLUSIVE = {'with_steps': True, 'with_terminal': True}


@pytest.mark.parametrize(
    'child_id, parent_id, kwargs, expected',
    [pytest.param(BOOK, None, FULL, TREE[BOOK],
                  id=f'child_id={BOOK}, include_self'),
     pytest.param(BOOK, None, EXCLUSIVE, [],
                  id=f'child_id={BOOK}'),
     pytest.param(BOOK, None, {}, [],
                  id=f'child_id={BOOK}'),
     pytest.param(WALI, None, FULL, TREE[WALI],
                  id=f'child_id={WALI}, include_self'),
     pytest.param(WALI, None, EXCLUSIVE, TREE[WALI][1:],
                  id=f'child_id={WALI}'),
     pytest.param(WALI, None, {}, [(c, p) for c, p, _, _ in TREE[WALI][1:]],
                  id=f'child_id={WALI}')])
def test_languoid_tree(treedb, child_id, parent_id, kwargs, expected):
    tree = treedb.Languoid.tree(**kwargs)

    select_tree = tree.select()

    if child_id is not None:
        select_tree = select_tree.where(tree.c.child_id == child_id)

    if parent_id is not None:
        select_tree = select_tree.where(tree.c.parent_id == parent_id)

    with treedb.connect() as conn:
        result = conn.execute(select_tree).all()

    assert result == expected


@pytest.mark.parametrize(
    'model, whereclause, expected_repr',
    [pytest.param(_models.Languoid, _models.Languoid.id == '3adt1234',
                  r"<Languoid id='3adt1234' level='dialect' name='3Ad-Tekles'>",
                  id='model=Languoid(id=3adt1234)'),
     pytest.param(_models.LanguoidLevel, _models.LanguoidLevel.name == 'language',
                  r"<LanguoidLevel name='language' description='[^']+' ordinal=\d+>",
                  id='model=LanguoidLevel(name=language)'),
     pytest.param(_models.PseudoFamily, _models.PseudoFamily.name == 'Bookkeeping',
                  r"<PseudoFamily languoid_id='book1242' name='Bookkeeping'"
                  r" config_section='bookkeeping' description='[^']+' bookkeeping=True>",
                  id='model=PseudoFamily(name=Bookkeeping)'),
     pytest.param(_models.Macroarea, _models.Macroarea.name == 'Eurasia',
                  r"<Macroarea name='Eurasia' config_section='[^']+' description='[^']+'>",
                  id='model=Marcoarea(name=Eurasia)'),
     pytest.param(_models.Country, _models.Country.id == 'RU',
                  r"<Country id='RU' name='Russian Federation'>",
                  id='model=Country(id=RU)'),
     pytest.param(_models.Link, _models.Link.title != sa.null(),
                  r"<Link languoid_id='\w+' ord=\d+ url='[^']+' title='[^']+' scheme='https?'>",
                  id='model=Link(title)'),
     pytest.param(_models.Timespan, None,
                  r"<Timespan start_year=\d+ start_month=\d+ start_day=\d+"
                  r" end_year=\d+ end_month=\d+ end_day=\d+>",
                  id='model=Timespan',
                  marks=pytest.mark.xfail_glottolog_tag('v4.1', reason='no timespan in Glottolog v4.1',
                                                        raises=AssertionError)),
     pytest.param(_models.Source, None,
                  r"<Source languoid_id='\w+' provider_id=\d+ bibitem_id=\d+>",
                  id='model=Source'),
     pytest.param(_models.SourceProvider, _models.SourceProvider.name == 'glottolog',
                  r"<SourceProvider name='glottolog'>",
                  id='model=SourceProvider(name=glottolog)'),
     pytest.param(_models.Bibfile, _models.Bibfile.name == 'hh',
                  r"<Bibfile id=\d+ name='hh'>",
                  id='model=Bibfile(name=hh)'),
     pytest.param(_models.Bibitem, None,
                  r"<Bibitem bibfile_id=\d+ bibkey='[^']+'>",
                  id='model=Bibitem'),
     pytest.param(_models.Altname, _models.Altname.lang != sa.null(),
                  r"<Altname languoid_id='\w+' provider_id=\d+ name='[^']+' lang='[^']*'>",
                  id='model=Altname(lang)'),
     pytest.param(_models.AltnameProvider, _models.AltnameProvider.name == 'multitree',
                  r"<AltnameProvider name='multitree'>",
                  id='model=AltnameProvider(name=multitree)'),
     pytest.param(_models.Trigger, None,
                  r"<Trigger languoid_id='\w+' field='[^']+' trigger='[^']+'>",
                  id='model=Trigger'),
     pytest.param(_models.Identifier, None,
                  r"<Identifier languoid_id='\w+' site_id=\d+ identifier='[^']+'>",
                  id='model=Identifier'),
     pytest.param(_models.IdentifierSite, _models.IdentifierSite.name == 'multitree',
                  "<IdentifierSite name='multitree'>",
                  id='model=IdentifierSite(name=multitree)'),
     pytest.param(_models.ClassificationComment, None,
                  r"<ClassificationComment languoid_id='\w+' kind='\w+' comment='[^']+'>",
                  id='model=ClassificatrionComment'),
     pytest.param(_models.ClassificationRef, None,
                  r"<ClassificationRef languoid_id='\w+' kind='\w+' bibitem_id=\d+>",
                  id='model=ClassificationRef'),
     pytest.param(_models.Endangerment, None,
                  r"<Endangerment languoid_id='\w+' status='[^']+'"
                  r" source_id=\d+ date=datetime\.datetime\([^)]+\)>",
                  id='model=Endangerment'),
     pytest.param(_models.EndangermentSource, _models.EndangermentSource.name == 'E22',
                  r"<EndangermentSource id=\d+ name='E22' bibitem_id=None pages=None>",
                  id='model=EndangermentSource(name=E22)'),
     pytest.param(_models.EthnologueComment, None,
                  r"<EthnologueComment languoid_id='\w+' isohid='[^']+' comment_type='[^']+'"
                  r" ethnologue_versions='[^']+'>",
                  id='model=EthnologueComment'),
     pytest.param(_models.IsoRetirement, None,
                  r"<IsoRetirement languoid_id='\w+' code='\w+' name='[^']+'"
                  r" change_request='[^']+' effective=datetime\.date\([^)]*\)"
                  r" reason='[^']*' remedy='[^']*'>",
                  id='model=IsoRetirement'),
     pytest.param(_models.IsoRetirementChangeTo, None,
                  r"<IsoRetirementChangeTo languoid_id='\w+' code='\w+'>",
                  id='model=IsoRetirementChangeTo')])
def test_repr(treedb, model, whereclause, expected_repr):
    query = sa.select(model)
    if whereclause is not None:
        query = query.where(whereclause)

    with treedb.Session() as session:
        inst = session.execute(query).scalars().first()

    result = repr(inst)

    assert re.fullmatch(expected_repr, result)



"""Fetch records from raw tables."""

from collections.abc import Iterator
import contextlib
import itertools
import logging

import sqlalchemy as sa

from .. import _globals
from .. import _tools
from .. import backend as _backend

from .models import File, Option, Value

__all__ = ['fetch_records']

WINDOWSIZE = 1_000


log = logging.getLogger(__name__)


def fetch_records(*, order_by: str = _globals.LANGUOID_ORDER,
                  progress_after: int = _tools.PROGRESS_AFTER,
                  windowsize: int = WINDOWSIZE,
                  skip_unknown: bool = True,
                  bind=_globals.ENGINE) -> Iterator[_globals.RecordItem]:
    """Yield (<path_part>, ...), <dict of <dicts of strings/string_lists>>) pairs."""
    try:
        dbapi_conn = bind.connection.driver_connection
    except AttributeError:
        dbapi_conn = None
    log.info('start generating raw records from %r', dbapi_conn or bind)

    # depend on no empty value files (save sa.outerjoin(File, Value) below)
    select_values = (sa.select(Value.file_id,
                               Option.section, Option.option, Option.is_lines,
                               Value.value)
                     .join_from(Value, Option))

    if order_by in ('file', True, None, False):
        key_column = File.id
        value_key = Value.file_id
    elif order_by in ('path', 'id'):
        select_values = select_values.join(File)
        if order_by == 'path':
            key_column = File.path
        else:
            key_column = File.glottocode
        value_key = key_column
    else:  # pragma: no cover
        raise ValueError(f'{order_by=!r} not implememted')

    log.info('order_by: %r', order_by)
    select_files = (sa.select(File.path)
                    .order_by(key_column))
    select_values = (select_values
                     .order_by(value_key,
                               'section', Value.line, 'option'))

    if skip_unknown:
        select_values = select_values.where(Option.is_lines != sa.null())

    groupby = (('file_id',), ('section',), ('option', 'is_lines'))
    groupby = itertools.starmap(_tools.groupby_attrgetter, groupby)
    groupby_file, groupby_section, groupby_option = groupby

    n = 0
    make_item = _globals.RecordItem.from_filepath_record
    with _backend.connect(bind=bind) as conn:
        for in_slice in window_slices(key_column, size=windowsize, bind=conn):
            if log.level <= logging.DEBUG:
                where = _backend.expression_compile(in_slice(key_column))
                log.debug('fetch rows %r', where.string)

            files = conn.execute(select_files.where(in_slice(key_column))).all()
            # single thread: no isolation level concerns
            values = conn.execute(select_values.where(in_slice(value_key))).all()
            # join in-memory by file_id total order index
            path_values = zip(files, groupby_file(values))

            count = 0
            for count, ((path,), (_, values)) in enumerate(path_values, start=1):
                record = {
                    s: {o: [ln.value for ln in lines] if is_lines else next(lines).value
                       for (o, is_lines), lines in groupby_option(sections)}
                    for s, sections in groupby_section(values)}
                yield make_item(path, record)

            assert count
            n += count
            if not (n % progress_after):
                log.info('%s raw records generated', f'{n:_d}')

    log.info('%s raw records total', f'{n:_d}')


def window_slices(key_column, /, *, size: int = WINDOWSIZE,
                  bind=_globals.ENGINE):
    """Yield where clause making function for key_column windows of size.

    adapted from https://github.com/sqlalchemy/sqlalchemy/wiki/RangeQuery-and-WindowedRangeQuery
    """
    log.info('fetch %r slices for window of %d', str(key_column.expression), size)
    keys = iterkeys(key_column, size=size, bind=bind)

    try:
        end = next(keys)
    except StopIteration:
        yield lambda c: sa.and_()
        return

    # right-inclusive indexes for windows of given size for continuous keys
    yield lambda c, end=end: (c <= end)
    last = end

    for end in keys:
        yield lambda c, last=last, end=end: sa.and_(c > last, c <= end)
        last = end

    yield lambda c, end=end: (c > end)


def iterkeys(key_column, /, *, size: int = WINDOWSIZE,
             bind=_globals.ENGINE):
    row_num = sa.func.row_number().over(order_by=key_column).label('row_num')
    select_all_keys = (sa.select(key_column.label('key'), row_num)
                       .alias('key_ord'))

    select_keys = (sa.select(select_all_keys.c.key)
                   .where((select_all_keys.c.row_num % size) == 0))

    log.debug('SELECT every %d-th %r using row_number() window function',
              size, str(key_column.expression))

    with (_backend.connect(bind=bind) as conn,
          contextlib.closing(conn.execute(select_keys)) as result):
        yield from result.scalars()



from helpers import assert_valid_languoids

import pytest


pytestmark = pytest.mark.raw


def test_iterlanguoids_from_raw(treedb, n=501):
    items = treedb.iterlanguoids('raw')
    assert_valid_languoids(items, n=n)



"""Raw ``.ini`` content as ``(path, section, option, line, value)`` rows."""

from .export import (print_stats,
                     checksum,
                     write_raw_csv,
                     write_files)

from .models import File, Option, Value
from .records import fetch_records

__all__ = ['print_stats',
           'checksum',
           'write_raw_csv',
           'write_files',
           'File', 'Option', 'Value',
           'fetch_records']



import pytest

from treedb import glottolog


pytestmark = pytest.mark.writes


@pytest.mark.xfail_glottolog_tag('master', reason='possibly unnormalized master',
                                 raises=AssertionError)
@pytest.mark.xfail_glottolog_tag('v5.2.1', reason='un-normalized v5.2.1 whitespace',
                                 # https://github.com/glottolog/glottolog/pull/1221
                                 raises=AssertionError)
@pytest.mark.xfail_glottolog_tag('v4.8', reason='extra v4.8 endangerment whitespace',
                                 # https://github.com/glottolog/glottolog/pull/981
                                 raises=AssertionError)
def test_roundtrip(treedb):
    if not glottolog.git_status_is_clean(treedb.root):
        raise RuntimeError(f'root must be clean for test: {treedb.root!r}')

    result = treedb.languoids.files.roundtrip()

    clean = glottolog.git_status_is_clean(treedb.root)
    if not clean:
        print(glottolog.git_status(treedb.root))

    if result is not None:
        raise RuntimeError(f'expected result is not None: {result!r}')
    assert clean



"""SQLite3 database engine."""

try:
    import sqlparse
except ImportError:  # pragma: no cover
    sqlparse = None

from ._basics import (print_versions,
                      set_engine, connect,
                      scalar, iterrows,
                      expression_compile,
                      json_object, json_datetime)

__all__ = ['sqlparse',
           'print_versions',
           'set_engine', 'connect',
           'scalar', 'iterrows',
           'expression_compile',
           'json_object',
           'json_datetime']



from helpers import (get_assert_head,
                     assert_nonempty_string,
                     assert_nonempty_string_tuple,
                     assert_nonempty_dict,
                     assert_valid_languoids)


def test_parse(bare_treedb, *, n=100):
    records = bare_treedb.languoids.iterrecords()

    items = bare_treedb.languoids.records.pipe(records, dump=False,
                                               convert_lines=True)

    assert_valid_languoids(items, n=n)


def test_dump(bare_treedb, *, n=100):
    languoids = bare_treedb.iterlanguoids()

    items = bare_treedb.languoids.records.pipe(languoids, dump=True,
                                               convert_lines=False)

    for path, record in get_assert_head(items, n=n):
        assert_nonempty_string_tuple(path)
        assert_nonempty_dict(record)

        assert_nonempty_string(record['core']['name'])

        assert record['core']['level'] in ('family', 'language', 'dialect')



"""Load Glottolog languoid model tables."""

import functools
import logging
import warnings

import sqlalchemy as sa

from .backend.models import Config
from .models import (LEVEL, SPECIAL_FAMILIES, BOOKKEEPING,
                     CLASSIFICATION,
                     Languoid, LanguoidLevel, PseudoFamily,
                     languoid_macroarea, Macroarea,
                     languoid_country, Country,
                     Link, Timespan,
                     Source, SourceProvider,
                     Bibfile, Bibitem,
                     Altname, AltnameProvider,
                     Trigger, Identifier, IdentifierSite,
                     ClassificationComment, ClassificationRef,
                     Endangerment, EndangermentStatus, EndangermentSource,
                     EthnologueComment,
                     IsoRetirement, IsoRetirementChangeTo)

__all__ = ['main']


log = logging.getLogger(__name__)


class ModelMap(dict):
    """Insert model (derive params f/ key), add ``key -> primary_key`` to map."""

    model = None

    @staticmethod
    def key_to_params(key):
        return {'name': key}

    def __init__(self, items=(), /, *,
                 conn,
                 model=None, key_to_params=None,
                 log_insert=True):
        super().__init__(items)
        if model is not None:
            self.model = model
        if key_to_params is not None:
            self.key_to_params = key_to_params
        self.log_insert = log_insert

        self.conn = conn
        self.insert = functools.partial(conn.execute, sa.insert(self.model))

    def __missing__(self, key, /):
        if self.log_insert:
            log.debug('insert new %s: %r', self.model.__tablename__, key)
        params = self.key_to_params(key)

        pk, = self.insert(params).inserted_primary_key

        self[key] = pk
        return pk


def main(languoids, /, *, conn):

    bibfile_ids = ModelMap(conn=conn, model=Bibfile)

    class BibitemMap(ModelMap):
        # using closure over bibfile_ids

        model = Bibitem

        @staticmethod
        def key_to_params(key):
            bibfile_name, bibkey = key
            return {'bibfile_id': bibfile_ids[bibfile_name], 'bibkey': bibkey}

        def pop_params(self, params):
            return self[params.pop('bibfile'), params.pop('bibkey')]

    bibitem_ids = BibitemMap(conn=conn,
                             log_insert=False)  # silence bibitems

    class EndangermentSourceMap(ModelMap):
        # using closure over bibitem_ids

        model = EndangermentSource

        @staticmethod
        def key_to_params(key):
            params = dict(key)
            if params.get('bibfile') is not None:
                params['bibitem_id'] = bibitem_ids.pop_params(params)
            return params

        @staticmethod
        def params_to_key(params):
            return tuple(sorted(params.items()))

    insert_languoid_levels(conn)

    insert_macroareas(conn)

    insert_endangermentstatus(conn, bibitem_ids=bibitem_ids)

    es_ids = EndangermentSourceMap(conn=conn)

    insert_languoids(conn,
                     languoids=languoids,
                     bibitem_ids=bibitem_ids, es_ids=es_ids)

    insert_pseudofamilies(conn)


def insert_languoid_levels(conn, /, *, config_file='languoid_levels.ini'):
    log.info('insert languoid levels from: %r', config_file)
    levels = Config.load(config_file, bind=conn)
    levels = dict(sorted(levels.items(), key=lambda x: int(x[1]['ordinal'])))

    assert set(levels) >= set(LEVEL)
    extra_levels = sorted(set(levels) - set(LEVEL))
    if extra_levels:
        warnings.warn(f'{config_file!r} has extra languoid levels:'
                      f' {extra_levels!r}')

    log.debug('insert %d languoid levels: %r', len(levels), list(levels))
    params = [{'name': section,
               'description': l['description'].strip(),
               'ordinal': int(l['ordinal'])}
              for section, l in levels.items()]
    conn.execute(sa.insert(LanguoidLevel), params)


def insert_pseudofamilies(conn, /, *, config_file='language_types.ini'):
    log.info('insert pseudofamilies from: %r', config_file)
    languagetypes = Config.load(config_file, bind=conn)
    pseudofamilies = {section: l for section, l in languagetypes.items()
                      if l.get('pseudo_family_id', '').strip()}

    log.debug('insert %d pseudofamilies: %r', len(pseudofamilies),
              list(pseudofamilies))
    params = [{'languoid_id': p['pseudo_family_id'].strip(),
               'name': p['category'].strip(),
               'config_section': section,
               'description': p.get('description', '').strip() or None,
               'bookkeeping': p['category'].strip() == BOOKKEEPING or None}
              for section, p in pseudofamilies.items()]
    conn.execute(sa.insert(PseudoFamily), params)

    query = sa.select(~sa.exists()
                      .where(PseudoFamily.languoid_id == Languoid.id)
                      .where(PseudoFamily.name != Languoid.name))
    assert conn.scalar(query), 'pseudo_family_id must be in-sync with category'

    inserted_special_families = {p['name'] for p in params if not p['bookkeeping']}
    unseen_special_families = inserted_special_families - set(SPECIAL_FAMILIES)
    missing_special_families = set(SPECIAL_FAMILIES) - inserted_special_families
    if unseen_special_families:
        warnings.warn(f'{config_file!r} has extra pseudofamilies:'
                      f' {unseen_special_families!r}')
    if missing_special_families:
        warnings.warn(f'{config_file!r} misses special families:'
                      f' {missing_special_families!r}')
    inserted_bookkeeping = {p['name'] for p in params if p['bookkeeping']}
    if inserted_bookkeeping != {BOOKKEEPING}:
        warnings.warn(f'inserted bookkeeping: {inserted_bookkeeping!r}'
                      f' expected bookkeeping: {BOOKKEEPING!r}')


def insert_macroareas(conn, /, *, config_file='macroareas.ini'):
    log.info('insert macroareas from: %r', config_file)
    macroareas = Config.load(config_file, bind=conn)

    log.debug('insert %d macroareas: %r', len(macroareas), list(macroareas))
    params = [{'name': m['name'].strip(),
               'config_section': section,
               'description': m['description'].strip()}
              for section, m in macroareas.items()]
    conn.execute(sa.insert(Macroarea), params)


def insert_endangermentstatus(conn, /, *, bibitem_ids,
                              config_file='aes_status.ini'):
    log.info('insert endangermentstatus from %r:', config_file)
    status = Config.load(config_file, bind=conn)

    log.debug('insert %d endangermentstatus: %r', len(status), list(status))
    params = [{'name': s['name'].strip(),
               'config_section': section,
               'ordinal': int(s['ordinal']),
               'egids': s['egids'].strip(),
               'unesco': s['unesco'].strip(),
               'elcat': s['elcat'].strip(),
               'icon': s['icon'].strip(),
               'bibitem_id': bibitem_ids[s['reference_id'].strip()
                                         .partition(':')[::2]]}
              for section, s in status.items()]
    conn.execute(sa.insert(EndangermentStatus), params)


def insert_languoids(conn, /, *, languoids, bibitem_ids, es_ids):
    log.info('insert languoids')

    def unseen_countries(countries, _seen={}):
        for c in countries:
            id_, name = (c[k] for k in ('id', 'name'))
            try:
                assert _seen[id_] == name
            except KeyError:
                _seen[id_] = name
                yield c

    kwargs = {'conn': conn,
              'insert_lang': functools.partial(conn.execute, sa.insert(Languoid)),
              'unseen_countries': unseen_countries,
              'sourceprovider_ids': ModelMap(conn=conn, model=SourceProvider),
              'bibitem_ids': bibitem_ids,
              'altnameprovider_ids': ModelMap(conn=conn, model=AltnameProvider),
              'identifiersite_ids': ModelMap(conn=conn, model=IdentifierSite),
              'es_ids': es_ids}

    for _, l in languoids:
        insert_languoid(l, **kwargs)


def insert_languoid(languoid, /, *, conn,
                    insert_lang,
                    unseen_countries,
                    sourceprovider_ids,
                    bibitem_ids,
                    altnameprovider_ids,
                    identifiersite_ids,
                    es_ids):
    macroareas = languoid.pop('macroareas')
    countries = languoid.pop('countries')
    links = languoid.pop('links')
    timespan = languoid.pop('timespan', None)

    sources = languoid.pop('sources', None)
    altnames = languoid.pop('altnames', None)
    triggers = languoid.pop('triggers', None)
    identifier = languoid.pop('identifier', None)
    classification = languoid.pop('classification', None)
    endangerment = languoid.pop('endangerment', None)
    hh_ethnologue_comment = languoid.pop('hh_ethnologue_comment', None)
    iso_retirement = languoid.pop('iso_retirement', None)

    lid = languoid['id']
    insert_lang(languoid)

    if macroareas:
        conn.execute(sa.insert(languoid_macroarea),
                     [{'languoid_id': lid, 'macroarea_name': ma}
                      for ma in macroareas])

    if countries:
        new_countries = list(unseen_countries(countries))
        if new_countries:
            ids = [n['id'] for n in new_countries]
            log.debug('insert new countries: %r', ids)

            conn.execute(sa.insert(Country), new_countries)

        conn.execute(sa.insert(languoid_country),
                     [{'languoid_id': lid, 'country_id': c['id']}
                      for c in countries])

    if links:
        conn.execute(sa.insert(Link),
                     [dict(languoid_id=lid, ord=i, **link)
                      for i, link in enumerate(links, start=1)])

    if timespan:
        conn.execute(sa.insert(Timespan),
                     dict(languoid_id=lid, **timespan))

    if sources is not None:
        for provider, data in sources.items():
            provider_id = sourceprovider_ids[provider]
            conn.execute(sa.insert(Source),
                         [dict(languoid_id=lid,
                               provider_id=provider_id,
                               bibitem_id=bibitem_ids.pop_params(s), **s)
                          for s in data])

    if altnames is not None:
        for provider, names in altnames.items():
            provider_id = altnameprovider_ids[provider]
            half, full = groups = ([], [])
            for n in names:
                r = dict(languoid_id=lid, provider_id=provider_id, **n)
                if 'lang' not in r:
                    half.append(r)
                elif not r['lang']:  # lang = r.get('lang') or server_default
                    r.pop('lang')
                    half.append(r)
                else:
                    full.append(r)
            for rows in groups:
                if rows:
                    conn.execute(sa.insert(Altname), rows)

    if triggers is not None:
        conn.execute(sa.insert(Trigger),
                     [{'languoid_id': lid, 'field': field,
                       'trigger': t, 'ord': i}
                      for field, triggers in triggers.items()
                      for i, t in enumerate(triggers, start=1)])

    if identifier is not None:
        conn.execute(sa.insert(Identifier),
                     [dict(languoid_id=lid,
                           site_id=identifiersite_ids[site],
                           identifier=i)
                      for site, i in identifier.items()])

    if classification is not None:
        for c, value in classification.items():
            isref, kind = CLASSIFICATION[c]
            if isref:
                conn.execute(sa.insert(ClassificationRef),
                             [dict(languoid_id=lid, kind=kind,
                                   bibitem_id=bibitem_ids.pop_params(r),
                                   ord=i, **r)
                              for i, r in enumerate(value, start=1)])
            else:
                conn.execute(sa.insert(ClassificationComment),
                             {'languoid_id': lid, 'kind': kind,
                              'comment': value})

    if endangerment is not None:
        source = es_ids.params_to_key(endangerment.pop('source'))
        conn.execute(sa.insert(Endangerment),
                     dict(languoid_id=lid, source_id=es_ids[source],
                          **endangerment))

    if hh_ethnologue_comment is not None:
        conn.execute(sa.insert(EthnologueComment),
                     dict(languoid_id=lid, **hh_ethnologue_comment))

    if iso_retirement is not None:
        change_to = iso_retirement.pop('change_to')
        conn.execute(sa.insert(IsoRetirement),
                     dict(languoid_id=lid, **iso_retirement))

        if change_to:
            conn.execute(sa.insert(IsoRetirementChangeTo),
                         [{'languoid_id': lid, 'code': c, 'ord': i}
                          for i, c in enumerate(change_to, start=1)])



import re

import pytest

import treedb


@pytest.mark.parametrize(
    'section, option, expected',
    [pytest.param('core', 'name', True,
                  id='section=core, option=name'),
     pytest.param('core', 'spam', False,
                  id='section=core, option=unknown'),
     pytest.param('spam', 'core', False,
                  id='section=unknown, option=unknown')])
def test_is_known(section, option, expected):
    assert treedb.languoids.fields.is_known(section, option) == expected


@pytest.mark.parametrize(
    'section, option, kwargs, expected',
    [pytest.param('core', 'name', {}, False,
                  id='section=core, option=name'),
     pytest.param('core', 'links', {}, True,
                  id='section=core, option=links'),
     pytest.param('core', 'WARNS_SCALAR', {}, (None, UserWarning, r'unknown'),
                  id='section=core, option=unknown'),
     pytest.param('core', 'RAISES_KEYERROR', {'unknown_as_scalar': False},
                  (KeyError, r'.+'),
                  id='section=core, option=unknown, strict')])
def test_is_lines(recwarn, section, option, kwargs, expected):
    if isinstance(expected, tuple):
        if len(expected) == 3:
            expected, warning, match = expected

            assert treedb.languoids.fields.is_lines(section, option, **kwargs) == expected

            w = recwarn.pop(warning)
            assert re.search(match, str(w.message))
        else:
            exception, match = expected

            with pytest.raises(exception, match=match):
                treedb.languoids.fields.is_lines(section, option, **kwargs)

            recwarn.pop(UserWarning)
    else:
        assert treedb.languoids.fields.is_lines(section, option, **kwargs) == expected

    assert not recwarn


@pytest.mark.parametrize(
    'section, options, expected',
    [pytest.param('core',
                  ['eggs', 'iso639-3', 'name'],
                  ['name', 'iso639-3', 'eggs'],
                  id='section=core, options=eggs-iso639-3-name')])
def test_sorted_options(section, options, expected):
    assert treedb.languoids.fields.sorted_options(section, options) == expected



"""Proxies for global default variables that can be updated."""

import datetime
import logging

import sqlalchemy as sa

from . import _tools

__all__ = ['PathProxy', 'SQLiteEngineProxy']


log = logging.getLogger(__name__)


class Proxy(object):

    _delegate = None

    def __getattr__(self, name):
        return getattr(self._delegate, name)

    def __repr__(self):
        return f'<{self.__module__}.{self.__class__.__name__}>'


class PathProxy(Proxy):
    """

    >>> PathProxy()
    <treedb._proxies.PathProxy>
    >>> assert PathProxy().inode() is None
    >>> import pathlib
    >>> PathProxy(pathlib.Path())
    <treedb._proxies.PathProxy path='.' inode=...>
    >>> print(PathProxy(pathlib.Path()))
    .
    """

    def __init__(self, path=None, /):
        self.path = path

    def __fspath__(self):
        return self._delegate.__fspath__()

    def __str__(self):
        if self._delegate is None:
            raise RuntimeError('str() on empty path proxy')
        return str(self._delegate)

    @property
    def path(self):
        return self._delegate

    @path.setter
    def path(self, path):
        if path is not None:
            path = _tools.path_from_filename(path)
        log.debug('set path of %r to %r', self, path)
        self._delegate = path

    def __repr__(self):
        if self.path is None:
            return super().__repr__()
        return (f'<{self.__module__}.{self.__class__.__name__}'
                f' path={self.path.as_posix()!r}'
                f' inode={self.inode()!r}>')

    def inode(self):
        if self.path is None or not self.path.exists():
            return None
        return self.path.stat().st_ino


class EngineProxy(Proxy, sa.engine.Engine):
    """

    >>> EngineProxy(future=True)
    <treedb._proxies.EngineProxy>
    """
    def __init__(self, engine=None, /, *, future):
        self.engine = engine
        self.future = future

    def _create_engine(self, url, /):
        log.debug('sqlalchemy.create_engine(%r)', url)
        self.engine = sa.create_engine(url, future=self.future)

    def connect(self, *, close_with_result: bool = False, **kwargs):
        return self.engine.connect(**kwargs)

    def dispose(self):
        if self._delegate is not None:
            log.debug('dispose %s.%r', self._delegate.__module__, self._delegate)
            self._delegate.dispose()

    @property
    def engine(self):
        return self._delegate

    @engine.setter
    def engine(self, engine):
        if engine is None:
            prefix = ''
        else:
            assert engine.url.drivername == 'sqlite'
            prefix = f'{engine.__module__}.'

        self.dispose()

        log.debug('set engine of %r to %s%r', self, prefix, engine)
        self._delegate = engine

    @property
    def url(self):
        return self.engine.url if self.engine is not None else None

    @url.setter
    def url(self, url):
        self._create_engine(url)

    def __repr__(self):
        if self.engine is None:
            return super().__repr__()
        return (f'<{self.__module__}.{self.__class__.__name__}'
                f' url={str(self.url)!r}>')


class SQLiteEngineProxy(EngineProxy):

    memory_write_path = None

    @property
    def file(self):
        if self.url is None or self.url.database is None:
            return None
        return _tools.path_from_filename(self.url.database)

    @file.setter
    def file(self, filename):
        url = 'sqlite://'
        if filename is not None:
            url += f'/{filename}'
        self.url = url

    def __repr__(self):
        if self.file is None:
            return super().__repr__()

        parent = ''
        name = self.file.as_posix()
        if self.file.is_absolute():
            parent = f' parent={self.file.parent.name!r}'
            name = self.file.name

        return (f'<{self.__module__}.{self.__class__.__name__}'
                f' filename={name!r}{parent}'
                f' size={self.file_size()!r}>')

    def file_with_suffix(self, suffix, /):
        if self.file is None:
            name = f'{self.memory_write_path.name}{suffix}'
            return self.memory_write_path.with_name(name)
        return self.file.with_suffix(suffix)

    def file_exists(self):
        return self.file is not None and self.file.exists()

    def file_mtime(self):
        return (datetime.datetime.fromtimestamp(self.file.stat().st_mtime)
                if self.file_exists() else None)

    def file_size(self, *, as_megabytes: bool = False):
        if self.file_exists():
            result = self.file.stat().st_size
            if as_megabytes:
                result /= 2**20
            return result
        return None

    def file_sha256(self):
        return _tools.sha256sum(self.file) if self.file_exists() else None



#!/usr/bin/env python3

"""Run code linting with https://flake8.pycqa.org."""

import pathlib
import sys

import flake8.main.cli


if __name__ == '__main__':
    print('run', [pathlib.Path(__file__).name] + sys.argv[1:])
    args = sys.argv[1:]

    # https://flake8.pycqa.org/en/latest/internal/cli.html#flake8.main.cli.main
    print(f'flake8.main.cli.main({args!r})')
    if (returncode := flake8.main.cli.main(args)):
        print('FAILED:', returncode)
        sys.exit(returncode)
    print('PASSED.')



"""Load and write ``glottolog/languoids/tree/**/md.ini``."""

from collections.abc import Iterable, Iterator
import functools
import logging
import os
from typing import NamedTuple
import warnings

from .. import _globals
from .. import _tools

from . import fields as _fields

__all__ = ['iterfiles',
           'roundtrip',
           'write_files']

BASENAME = _globals.LANGUOID_FILE_BASENAME


log = logging.getLogger(__name__)


class ConfigParser(_tools.ConfigParser):
    """ConfigParser for ``glottolog/languoids/tree/**/md.ini``."""

    _basename = BASENAME

    _newline = '\r\n'

    _header = '# -*- coding: {encoding} -*-\n'

    def update_config(self, raw_record: _fields.RawRecordType, /, *,
                      replace: bool = False,
                      quiet: bool = False,
                      is_lines=_fields.is_lines,
                      core_sections=_fields.CORE_SECTIONS,
                      omit_empty_core_options=_fields.OMIT_EMPTY_CORE_OPTIONS,
                      keep_empty_sections=_fields.KEEP_EMPTY_SECTIONS,
                      keep_empty_options=_fields.KEEP_EMPTY_OPTIONS,
                      sorted_sections=_fields.sorted_sections,
                      sorted_options=_fields.sorted_options) -> bool:
        if replace:
            self.clear()

        for core_section in core_sections:
            s = raw_record[core_section]
            for core_option in omit_empty_core_options:
                if core_option in s and not s[core_option]:
                    del s[core_option]

        changed = False

        old_sections = set(self.sections())
        old_empty = {s for s in old_sections if not any(v for _, v in self.items(s))}

        new_sections = {sec for sec, s in raw_record.items() if s}
        leave = (old_empty - new_sections) & keep_empty_sections

        drop = old_sections - core_sections - leave - new_sections
        if drop:
            drop = sorted_sections(drop)
            log.debug('cfg.remove_section(s) for s in %r', drop)
            for s in drop:
                self.remove_section(s)
            changed = True

        add = new_sections - old_sections
        if add:
            add = sorted_sections(add)
            if not quiet:
                log.debug('cfg.add_section(s) for s in %r', add)
            for s in add:
                self.add_section(s)
            changed = True

        for section in sorted_sections(raw_record):
            if section in drop:
                continue

            s = raw_record[section]

            if section not in old_sections or section in core_sections:
                pass
            elif section in leave:
                continue
            else:
                old_options = {o for o in self.options(section)
                               if (section, o) not in keep_empty_options}
                new_options = {k for k, v in s.items() if v}
                drop_options = old_options - new_options

                if drop_options:
                    drop_options = sorted_options(section, drop_options)
                    log.debug('cfg.remove_option(%r, o) for o in %r',
                              section, drop_options)
                    for o in drop_options:
                        self.remove_option(section, o)
                    changed = True

            for option in sorted_options(section, s):
                value = s[option]
                if value is None or (not value and is_lines(section, option)):
                    continue

                old = self.get(section, option, fallback=None)
                if old == value:
                    continue

                if not quiet:  # pragma: no cover
                    if old is None:
                        log.debug('cfg add option (%r, %r)', section, option)
                    log.debug('cfg.set_option(%r, %r, %r)', section, option, value)

                self.set(section, option, value)
                changed = True

        return changed


class FileInfo(NamedTuple):
    """Triple of ((<path_part>, ...), <DirEntry object>, <ConfigParser object>)."""

    path: _globals.PathType

    dentry: os.DirEntry

    config: ConfigParser

    @classmethod
    def from_dentry(cls, dentry: os.DirEntry, /, *,
                    path_slice: slice = slice(None)):
        path = _tools.path_from_filename(dentry)
        config = ConfigParser.from_file(path)
        return cls(path.parts[path_slice], dentry, config)


def iterfiles(root=_globals.ROOT, /, *,
              progress_after: int = _tools.PROGRESS_AFTER
              ) -> Iterator[FileInfo]:
    """Yield triples of ((<path_part>, ...), <ConfigParser object>, <DirEntry object>)."""
    root = _tools.path_from_filename(root).resolve()
    log.info(f'start parsing {BASENAME} files from %r', root)
    msg = f'%s {BASENAME} files parsed'

    kwargs = {'path_slice': slice(len(root.parts), -1)}
    make_fileinfo = functools.partial(FileInfo.from_dentry, **kwargs)

    n = 0
    for n, dentry in enumerate(_tools.walk_scandir(root), start=1):
        yield make_fileinfo(dentry)

        if not (n % progress_after):
            log.info(msg, f'{n:_d}')

    log.info(f'%s {BASENAME} files total', f'{n:_d}')


def roundtrip(root=_globals.ROOT, /, *,
              progress_after: int = _tools.PROGRESS_AFTER) -> None:
    """Load/save all config files (drops leading/trailing whitespace)."""
    log.info(f'start roundtripping {BASENAME} files in %r', root)
    for path_tuple, dentry, cfg in iterfiles(root, progress_after=progress_after):
        cfg.to_file(dentry.path)


def write_files(records: Iterable[_globals.RecordItem], /,
                root=_globals.ROOT, *, replace: bool = False,
                dry_run: bool = False, quiet: bool | None = None,
                require_nwritten: int | None = None,
                progress_after: int | None = _tools.PROGRESS_AFTER,
                basename: str = BASENAME) -> int:
    """Write ((<path_part>, ...), <dict of dicts>) pairs to root."""
    if replace:  # pragma: no cover
        if dry_run:
            warnings.warn('replace=True ignored by dry_run=True')
        else:
            log.warning(f'replace present {basename} files')

    if quiet is None:
        quiet = dry_run or replace

    root = _tools.path_from_filename(root)
    log.info(f'start writing {basename} files into %r', root)

    load_config = ConfigParser.from_file

    files_written = 0

    for path_tuple, raw_record in map(_fields.join_lines_inplace, records):
        path = root.joinpath(*path_tuple + (basename,))
        cfg = load_config(path)
        changed = cfg.update_config(raw_record, replace=replace, quiet=quiet)

        if changed:
            if not dry_run:
                if not replace:
                    log.info('write cfg.to_file(%r)', path)
                cfg.to_file(path)

            files_written += 1

            if not dry_run and (files_written % progress_after):
                log.info(f'%s {basename} files written', f'{files_written:_d}')

        if require_nwritten is not None and files_written > require_nwritten:
            raise ValueError(f'{files_written=} over {require_nwritten=}')

    log.info(f'%s {basename} files written total', f'{files_written:_d}')
    if require_nwritten is not None and files_written < require_nwritten:
        raise ValueError(f'{files_written=} under {require_nwritten=}')
    return files_written



"""Insert raw model tables."""

import functools
import logging

import sqlalchemy as sa

from .. import _globals
from .. import _tools
from .. import languoids as _languoids
from ..languoids import fields as _fields

from .models import File, Option, Value

__all__ = ['main']


log = logging.getLogger(__name__)


class OptionMap(dict):
    """Insert option on demand, add ``(section, option) -> (pk, is_lines)`` to map."""

    model = Option

    def __init__(self, items=(), /, *, conn):
        super().__init__(items)
        self.insert = functools.partial(conn.execute, sa.insert(self.model))

    def __missing__(self, key, /):
        log.debug('insert option: %r', key)
        section, option = key
        is_lines = _fields.is_lines(section, option)

        params = {'section': section, 'option': option,
                  'is_lines': is_lines}

        ord_section = _fields.SECTION_ORDER.get(section)
        if ord_section is not None and _fields.is_all_options(*key):
            ord_option = 0
            params.update(ord_option=ord_option,
                          defined=True,
                          defined_any_options=True)
        else:
            ord_option = _fields.FIELD_ORDER.get(key)
            params.update(ord_option=ord_option,
                          defined=ord_option is not None,
                          defined_any_options=False)

        params.update(ord_section=ord_section,
                      ord_option=ord_option)

        pk, = self.insert(params).inserted_primary_key

        self[key] = result = (pk, is_lines)
        return result


def itervalues(cfg, file_id, /, *, option_map):
    get_line = _tools.next_count(start=1)
    for section, sec in cfg.items():
        for option, text in sec.items():
            option_id, is_lines = option_map[section, option]
            values = text.strip().splitlines() if is_lines else [text]
            for v in values:
                yield {'file_id': file_id, 'option_id': option_id,
                       'line': get_line(), 'value': v}


def main(root, /, *, conn):
    insert_file = functools.partial(conn.execute, sa.insert(File))

    option_id_is_lines = OptionMap(conn=conn)

    insert_value = functools.partial(conn.execute, sa.insert(Value))

    for path_tuple, dentry, cfg in _languoids.iterfiles(root):
        sha256 = _tools.sha256sum(dentry.path, raw=True)

        file_params = {'glottocode': path_tuple[-1],
                       'path': _globals.FILE_PATH_SEP.join(path_tuple),
                       'size': dentry.stat().st_size,
                       'sha256': sha256.hexdigest()}
        file_id, = insert_file(file_params).inserted_primary_key

        value_params = list(itervalues(cfg, file_id, option_map=option_id_is_lines))

        insert_value(value_params)



import itertools

MB = 2**20


def pairwise(iterable, /):
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


def get_assert_head(items, /, *, n):
    head = list(itertools.islice(items, n))

    assert head
    assert len(head) == n

    return head


def assert_nonempty_string(obj, /):
    assert obj is not None
    assert isinstance(obj, str)


def assert_nonempty_string_tuple(obj, /):
    assert obj is not None
    assert isinstance(obj, tuple)
    assert all(isinstance(o, str) for o in obj)
    assert obj
    assert all(obj)


def assert_nonempty_dict(obj, /):
    assert obj is not None
    assert isinstance(obj, dict)
    assert obj


def assert_file_size_between(path, /, min, max, *, unit=MB):
    assert path is not None
    assert path.exists()
    assert path.is_file()
    assert min * unit <= path.stat().st_size <= max * unit


def assert_valid_languoids(items, /, *, n):
    for path, languoid in get_assert_head(items, n=n):
        assert_nonempty_string_tuple(path)
        assert_nonempty_dict(languoid)

        for key in ('id', 'level', 'name'):
            assert_nonempty_string(languoid[key])

        assert languoid['parent_id'] or languoid['parent_id'] is None

        assert languoid['level'] in ('family', 'language', 'dialect')



"""Set the root for the ``languoids/tree/**/md.ini`` file tree."""

import logging

from .. import _globals
from .. import _tools

__all__ = ['set_root', 'get_repo_root']

TREE_IN_ROOT = _tools.path_from_filename('languoids', 'tree')


log = logging.getLogger(__name__)


def set_root(repo_root, /, *, resolve: bool = False,
             treepath=TREE_IN_ROOT):
    """Set and return default root for glottolog lanugoid directory tree."""
    log.info('set_root: %r', repo_root)
    if repo_root is None:
        raise ValueError(f'missing repo_root path: {repo_root!r}')

    repo_path = _tools.path_from_filename(repo_root)
    if resolve:
        repo_path = repo_path.resolve(strict=False)

    ROOT = _globals.ROOT  # noqa: N806
    ROOT.path = repo_path / _tools.path_from_filename(treepath)
    return ROOT


def get_repo_root(root=_globals.ROOT, /, *,
                  treepath=TREE_IN_ROOT):
    assert root.parts[-len(treepath.parts):] == treepath.parts
    repo_root = _tools.path_from_filename(root)
    for _ in treepath.parts:
        repo_root = repo_root.parent
    return repo_root



import re

import pytest


@pytest.mark.parametrize(
    'table, expected',
    [pytest.param('__dataset__',
                  re.compile(r'CREATE TABLE __dataset__ \(\n'
                             r'.+\n'
                             r'0', re.DOTALL),
                  id='table=__dataset__')])
def test_print_table_sql(capsys, empty_treedb, table, expected):
    assert empty_treedb.print_table_sql(table) is None

    out, err = capsys.readouterr()
    assert not err

    assert expected.fullmatch(out.strip())


def test_select_tables_nrows(empty_treedb):
    query = empty_treedb.select_tables_nrows()
    with empty_treedb.connect() as conn:
        rows = conn.execute(query).all()

    assert rows

    for table, nrows in rows:
        assert isinstance(table, str)
        assert table

        assert isinstance(nrows, int)
        assert nrows == 0


def test_select_views(empty_treedb):
    query = empty_treedb.backend.sqlite_master.select_views()

    with empty_treedb.connect() as conn:
        names = conn.execute(query).scalars().all()

    assert names == ['example', 'path_languoid', 'stats']



#!/usr/bin/env python3
# flake8: noqa

"""Fix double whitespace in classification > sub."""

import sqlalchemy as sa
import treedb


treedb.configure_logging(level='INFO')

treedb.set_root('../glottolog')

treedb.set_engine('treedb.sqlite3')

treedb.load()

treedb.check()

update = (sa.update(treedb.raw.Value)
          .where(sa.select(1)
                 .where(treedb.raw.Value.option_id == treedb.raw.Option.id)
                 .where(treedb.raw.Option.section == 'classification')
                 .where(treedb.raw.Option.option.in_(['sub', 'family']))
                 .exists())
          .where(treedb.raw.Value.value.like('%  %'))
          .values(value=sa.func.replace(treedb.raw.Value.value, '  ', ' ')))

print(update)

with treedb.connect() as conn:
    result = conn.execute(update)
    conn.commit()

print('', f'{result.rowcount:d} lines replaced', sep='\n')  # 42, 0

#print('', f"{treedb.write_files(source='raw')} files written", sep='\n')  # 42, 0
